import sys
import inspect
import modules.general
import modules.blender
import modules.terrain
import modules.modifiers
import modules.generating
import modules.generating_function
import modules.combining
import modules.other
import modules.loop


MODULES = {}


# ## Add from all module files ###
# data
clsmembers_pairs = inspect.getmembers(
    sys.modules[modules.general.__name__], inspect.isclass)
for (k, v) in clsmembers_pairs:
    MODULES[k] = v

# modifiers
clsmembers_pairs = inspect.getmembers(
    sys.modules[modules.modifiers.__name__], inspect.isclass)
for (k, v) in clsmembers_pairs:
    MODULES[k] = v

# generating
clsmembers_pairs = inspect.getmembers(
    sys.modules[modules.generating.__name__], inspect.isclass)
for (k, v) in clsmembers_pairs:
    MODULES[k] = v

# generating_function
clsmembers_pairs = inspect.getmembers(
    sys.modules[modules.generating_function.__name__], inspect.isclass)
for (k, v) in clsmembers_pairs:
    MODULES[k] = v

# combining
clsmembers_pairs = inspect.getmembers(
    sys.modules[modules.combining.__name__], inspect.isclass)
for (k, v) in clsmembers_pairs:
    MODULES[k] = v

# other
clsmembers_pairs = inspect.getmembers(
    sys.modules[modules.other.__name__], inspect.isclass)
for (k, v) in clsmembers_pairs:
    MODULES[k] = v

# loop over pipes
clsmembers_pairs = inspect.getmembers(
    sys.modules[modules.loop.__name__], inspect.isclass)
for (k, v) in clsmembers_pairs:
    MODULES[k] = v

# blender
clsmembers_pairs = inspect.getmembers(
    sys.modules[modules.blender.__name__], inspect.isclass)
for (k, v) in clsmembers_pairs:
    MODULES[k] = v

# terrain
clsmembers_pairs = inspect.getmembers(
    sys.modules[modules.terrain.__name__], inspect.isclass)
for (k, v) in clsmembers_pairs:
    MODULES[k] = v



import sys
import subprocess


def get_installed_packages():
    try:
        # Using subprocess to get the list of installed packages
        result = subprocess.run([sys.executable, '-m', 'pip', 'list'],
                                capture_output=True, text=True)
        return result.stdout
    except Exception as e:
        return f"Error retrieving installed packages: {e}"


def main():
    # Print Python version
    python_version = sys.version
    print("Python Version:")
    print(python_version)
    print()

    # Print path to the Python executable
    python_executable = sys.executable
    print("Path to Python Executable:")
    print(python_executable)
    print()

    # Print installed packages
    print("Installed Packages:")
    installed_packages = get_installed_packages()
    print(installed_packages)


if __name__ == "__main__":
    main()



import numpy as np
import matplotlib
import colorcet as cc
import matplotlib.pyplot as plt
from matplotlib.backends.backend_agg import FigureCanvasAgg as FigureCanvas
from matplotlib.figure import Figure
import matplotlib.cm as cm
matplotlib.pyplot.switch_backend('Agg')  # To run on keb
import matplotlib.ticker as plticker


def new_fig(nrows=1, ncols=1, **kwargs):
    """Create a new matplotlib figure containing one axis

    Return:
      (fig, ax) or (fig, list[ax])
    """
    fig = Figure(**kwargs)
    FigureCanvas(fig)
    if nrows == 1 and ncols == 1:
        # Create a single axis
        ax = fig.add_subplot(1, 1, 1)
        return fig, ax
    else:
        # Create a list of axes
        axs = [fig.add_subplot(nrows, ncols, index) for index in
               range(1, nrows*ncols+1)]
        return fig, axs


def new_3d_fig(**kwargs):
    """ Test 3D subplot """
    fig = Figure(**kwargs)
    FigureCanvas(fig)
    ax = fig.add_subplot(111, projection='3d')

    return fig, ax


def save_all_axes(fig, base_filename, ticks_in=True, delta=0.05,
                  dx=None, dy=None, clear_legend=False, axis_off=False,
                  dpi=None, **savefig_kwargs):
    '''
    Save all axes as separate images.

    Args:
      ...
      delta:
      dx: explicity delta in x
      dy: explicity delta in y
    '''
    for i, ax in enumerate(fig.get_axes()):
        if ticks_in:
            ax.tick_params(direction="in", which='both')

        if axis_off:
            ax.axis('off')

        # Print limits
        print(f"#### {base_filename} ####")
        print(f"  (ax.get_xlim(), ax.get_ylim()):{(ax.get_xlim(), ax.get_ylim())}")

        if clear_legend:
            legend = ax.get_legend()
            if legend:
                for text in legend.get_texts():
                    text.set_text(' '*int(1.5*len(text.get_text())))

        filename = base_filename.replace('.png', '_{:01d}.png'.format(i))
        save_subfig(fig, ax, filename, delta=delta, dpi=dpi, **savefig_kwargs)


def strip_axis(ax):
    '''
    Remove all labels etc. from axis
    '''
    ax.set_yticklabels([])
    ax.set_xticklabels([])
    ax.set_xlabel("")
    ax.set_ylabel("")
    ax.set_title("")
    try:
        ax.get_legend().remove()
    except AttributeError:
        pass


def strip_figure(fig):
    '''
    Remove all visible text from figure
    '''
    for txt in fig.texts:
        txt.set_visible(False)
        txt.set_text("")


def save_subfig(fig, ax, filename, delta=0.05, extent=None, dx=None, dy=None,
                dpi=None, **savefig_kwargs):
    '''
    Save subfigure containing axis
    '''
    if extent is None:
        extent = get_bbox(fig, ax, delta, dx, dy)
        print("extent:{}".format(extent))
        print("extent.width:{}".format(extent.width))
        print("extent.height:{}".format(extent.height))
    fig.savefig(filename, bbox_inches=extent, dpi=dpi, **savefig_kwargs)


def get_bbox(fig, ax, delta=0.05, dx=None, dy=None):
    '''
    Calculate bounding box around axis
    '''
    extent = ax.get_window_extent().transformed(
        fig.dpi_scale_trans.inverted())
    # get points
    p = extent.get_points()
    # Calculate new extent
    x0 = p[0, 0]
    y0 = p[0, 1]
    w = p[1, 0] - x0
    h = p[1, 1] - y0
    # d = delta
    dx = delta if dx is None else dx
    dy = delta if dy is None else dy
    extent = extent.from_bounds(x0 - dx, y0 - dy, w + 2 * dx, h + 2 * dy)

    return extent


def add_right_panel(fig, ax):
    [[x0, y0], [x1, y1]] = ax.get_position().get_points()
    left, width = x0, x1-x0
    bottom, height = y0, y1-y0
    spacing = 0.02
    margin_height = 0.07
    rect = [left + width + spacing, bottom, margin_height, height]

    ax_histy = fig.add_axes(rect, sharey=ax)

    # Trick to turn of tick-labels just on marginal axis
    plt.setp(ax_histy.get_yticklabels(), visible=False)

    return ax_histy


def add_right_histogram(ax_right, values, barh_kwargs={}, hist_kwargs={}):
    ''' Given an axis, to the right of the main canvas, with 'share-y' '''
    hist, bin_edges = np.histogram(values, **hist_kwargs)
    bins = bin_edges[:-1] + (bin_edges[1] - bin_edges[0]) / 2
    height = 0.9 * (bin_edges[1] - bin_edges[0])
    barh = ax_right.barh(y=bins, width=hist, height=height, **barh_kwargs)

    return barh


def add_multicolored_line(
        ax, y, x=None, colors=None, color_values=None,
        linewidth=2, cmap=None,
):
    ''' Add plot to axis, with multiple colors

    Each line consists of multiple segments, and these are all
    colored differently

    Args:
      ax: (plt axis)
      y: (np.array) of shape (N,)
      x: (np.array) of shape (N,), or None (-> infered from y)
      colors: (np.array) of shape (N, 4), or None
      color_values: (np.array) of shape (N,) or None
      cmap: None or matplotlib.cm

    Return:
      ax, line

    '''
    # From
    # https://matplotlib.org/stable/gallery/lines_bars_and_markers/multicolored_line.html

    from matplotlib.collections import LineCollection

    # Infer x from y, if it is not given
    if x is None:
        x = np.linspace(0, y.shape[0], y.shape[0])

    # Create a set of line segments so that we can color them individually
    # This creates the points as a N x 1 x 2 array so that we can stack points
    # together easily to get the segments. The segments array for line collection
    # needs to be (numlines) x (points per line) x 2 (for x and y)
    points = np.array([x, y]).T.reshape(-1, 1, 2)
    segments = np.concatenate([points[:-1], points[1:]], axis=1)

    # Create 'LineCollection'
    # Set colors either from 4-vector or 1-vector and cmap
    if colors is not None:
        lc = LineCollection(segments, colors=colors)
    else:
        lc = LineCollection(segments, cmap=cmap)
        # Use color_values. Has to be 1D vector
        lc.set_array(color_values.squeeze())

    lc.set_linewidth(linewidth)

    line = ax.add_collection(lc)

    return ax, line


def add_4m_grid(ax, **_):
    # loc = plticker.MultipleLocator(base=4.0)
    # ax.xaxis.set_minor_locator(loc)
    # ax.yaxis.set_minor_locator(loc)
    # ax.grid(color='k', which='major')
    # ax.grid(color='k', which='minor', linewidth=0.25)
    add_grid(ax, size=4)
    add_grid(ax, size=20, linewidth=0.5)


def add_grid(ax, size=10, linewidth=0.25, color='k', **kwargs):
    '''
    Add a grid

    Args:
      size (float or [float, float])
    '''
    # TODO: Does not handle left/bottom limits
    xlim = ax.get_xlim()
    ylim = ax.get_ylim()

    xmax = xlim[1]
    ymax = ylim[1]

    if isinstance(size, (float, int)):
        size_x = size
        size_y = size
    elif isinstance(size, list):
        size_x, size_y = size

    vlines = np.arange(0, xmax, size_x)
    hlines = np.arange(0, ymax, size_y)

    # Add to kwargs
    kwargs['linewidth'] = linewidth
    kwargs['color'] = color

    ax.vlines(vlines, 0, ymax, **kwargs)
    ax.hlines(hlines, 0, xmax, **kwargs)


def plot_obstacles(
        obstacles,
        filename=None, fig_ax=None,
        xlim=None, ylim=None,
        dpi=400, color='radius',
        cmap=cc.cm.rainbow,
        cbar_label='Height (m)',
        symmetric_vlims=True,
):
    '''
    Plot obstacles from arrays of position and radius

    Args:

      color: str or np.array
    '''
    radius = obstacles.width/2
    position = obstacles.position

    if fig_ax is not None:
        fig, ax = fig_ax
    else:
        fig, ax = new_fig()

    # If no limits are provided, and ax-limits have default values
    # Estimate limits
    # print(f"ax.get_autoscalex_on():{ax.get_autoscalex_on()}")
    if xlim is None and ax.get_xlim() == (0, 1):
        xlim = [np.min(position[0, :]), np.max(position[0, :])]
    if ylim is None and ax.get_ylim() == (0, 1):
        ylim = [np.min(position[1, :]), np.max(position[1, :])]

    if xlim is not None:
        ax.set_xlim(xlim)
    if ylim is not None:
        ax.set_ylim(ylim)

    if isinstance(color, str) and color == 'radius':
        c = radius/np.max(radius)
    else:
        c = color

    kwargs = {}
    if symmetric_vlims:
        vbound = max(np.max(color), -np.min(color))  # ensures 0 is in the center
        vmin, vmax = -vbound, vbound
        kwargs['vmin'] = vmin
        kwargs['vmax'] = vmax

    # Draw canvas to be able to get size
    xlim = ax.get_xlim()
    ax.set_aspect(1)
    ax_width = xlim[1] - xlim[0]
    fig.canvas.draw()

    # Plot points
    factor = ((2 * ax.get_window_extent().width/ax_width * 72./fig.dpi) ** 2)
    s = factor * radius**2
    sc = ax.scatter(*position, edgecolor='none', s=s, cmap=cmap, c=c, alpha=0.9, **kwargs)
    fig.colorbar(sc, ax=ax, label=cbar_label)

    if filename is not None:
        fig.savefig(filename, dpi=dpi)
    else:
        return fig, ax


def plot_image(
        array_2d, filename=None, fig_ax=None,
        extent=None, cmap=cc.cm.gray, xlim=None, ylim=None, **imshow_kwargs):
    '''
    Plot some image...

    Args:
      extent=None or [xmin, xmax, ymin, ymax],
    '''
    if fig_ax is not None:
        fig, ax = fig_ax
    else:
        fig, ax = new_fig()

    # Set limits, unless already set
    # Use extent if given, or else size of array
    if xlim is None:
        if extent is None:
            xlim = [0, array_2d.shape[0]]
        else:
            xlim = extent[:2]
    if ylim is None:
        if extent is None:
            ylim = [0, array_2d.shape[1]]
        else:
            ylim = extent[2:]

    # Set extent unless it has already been set
    # As default is [-0.5, numcols-0.5] we have to set this
    if extent is None:
        extent = [xlim[0], xlim[1], ylim[0], ylim[1]]

    # Set limits
    ax.set_xlim(xlim)
    ax.set_ylim(ylim)

    im = ax.imshow(
            array_2d.T, extent=extent,
            cmap=cmap, origin='lower', **imshow_kwargs)
    fig.colorbar(im)

    fig.tight_layout()
    if filename is not None:
        fig.savefig(filename)
    else:
        return fig, ax


def plot_probability(array_2d, filename=None, cmap=cc.cm.gray, vmin=0, **kwargs):
    return plot_image(array_2d, cmap=cmap, filename=filename, vmin=vmin, **kwargs)


def plot_diverging(array_2d, filename=None, cmap=cc.cm.coolwarm, vmin=-1, vmax=1, **kwargs):
    return plot_image(
        array_2d,
        filename=filename,
        cmap=cmap,
        vmin=vmin,
        vmax=vmax,
        **kwargs)


def plot_y_array(array_2d, grid, filename=None, cmap=cc.cm.blues, vmin=1, vmax=5, **kwargs):
    return plot_image(
        array_2d,
        filename=filename,
        cmap=cmap,
        vmin=vmin,
        vmax=vmax,
        extent=grid.extent,
        **kwargs)


def plot_histogram(array, bins=10, range=None, filename=None):
    # Construct histogram
    hist, bin_edges = np.histogram(array, bins, range)
    # Plot histogram
    fig, ax = new_fig()
    ax.bar(np.arange(bins)+1, hist)

    if filename is not None:
        fig.savefig(filename)
    else:
        return fig, ax


def plot_hf(hf_array, extent, ax=None, fig=None,
            label='Height (m)', **_):
    '''
    plot hf
    '''
    if ax is None:
        fig, ax = new_fig()

    # Plot image
    im = ax.imshow(hf_array.T, origin='lower', extent=extent)
    fig.colorbar(im, label=label)

    return fig, ax


def plot_terrain(terrain, ax=None, fig=None,
                 cmap=cc.cm.dimgray,
                 label='Height (m)', **kwargs):
    '''
    plot hf
    '''
    if ax is None:
        fig, ax = new_fig()

    # Plot image
    im = ax.imshow(terrain.array.T, origin='lower', extent=terrain.extent,
                   cmap=cmap, **kwargs)
    fig.colorbar(im, label=label)

    return fig, ax


def debug_plot(filename=None, terrain_temp=[], terrain_prim=[], **kwargs):
    temp_length = len(terrain_temp)
    heap_length = len(terrain_prim)

    max_length = max(temp_length, heap_length)

    if max_length == 0:
        return

    fig, axs = new_fig(nrows=max_length, ncols=2, figsize=(2.65*3, max_length*3))
    axs = np.array(axs).reshape(max_length, 2)

    # Plot heap
    for i, terrain in enumerate(terrain_prim):
        ax = axs[i, 0]
        fig, ax = plot_terrain(terrain, ax=ax, fig=fig)
        ax.patch.set_edgecolor('red')
        ax.patch.set_linewidth(10)

    # Plot temp
    for i, terrain in enumerate(terrain_temp):
        ax = axs[i, 1]
        fig, ax = plot_terrain(terrain, ax=ax, fig=fig)
        ax.patch.set_edgecolor('green')
        ax.patch.set_linewidth(10)

    # turn of unused axes
    for ax in axs.flatten():
        if not ax.images:
            ax.set_axis_off()

    # Save
    fig.tight_layout()
    fig.savefig(filename)


def debug_plot_horizontal(
        filename=None, terrain_temp=[],
        terrain_prim=[], **kwargs):
    temp_length = len(terrain_temp)
    heap_length = len(terrain_prim)

    max_length = max(temp_length, heap_length)

    if max_length == 0:
        return

    fig, axs = new_fig(nrows=2, ncols=max_length, figsize=(max_length*3, 3*1.45))
    axs = np.array(axs).reshape(2, max_length)

    # Plot heap
    for i, terrain in enumerate(terrain_prim):
        ax = axs[0, i]
        fig, ax = plot_terrain(terrain, ax=ax, fig=fig)
        ax.patch.set_edgecolor('red')
        ax.patch.set_linewidth(10)

    # Plot temp
    for i, terrain in enumerate(terrain_temp):
        ax = axs[1, i]
        fig, ax = plot_terrain(terrain, ax=ax, fig=fig)
        ax.patch.set_edgecolor('green')
        ax.patch.set_linewidth(10)

    # turn of unused axes
    for ax in axs.flatten():
        if not ax.images:
            ax.set_axis_off()

    # Save
    fig.tight_layout()
    fig.savefig(filename)


def generate_line_grid(size=(512, 512),
                       extent=(-50, 50, -50, 50),
                       major_step=100,
                       medium_step=10,
                       minor_step=1,
                       filename=None,
):
    """
    Generate a grid using explicit line drawing.

    Args:
        size (tuple): (width, height) in pixels.
        extent (tuple): [xmin, xmax, ymin, ymax] in 'meters' (or units).
        major_step (float): Spacing of thick lines.
        medium_step (float): Spacing of medium lines.
        minor_step (float): Spacing of thin lines.
        filename (str): Optional file name to save.
    """
    width_px, height_px = size
    xmin, xmax, ymin, ymax = extent

    dpi = 100
    fig, ax = new_fig(figsize=(width_px / dpi, height_px / dpi), dpi=dpi)

    # Set custom limits
    ax.set_xlim(xmin, xmax)
    ax.set_ylim(ymin, ymax)

    # Background
    fig.patch.set_facecolor("white")
    ax.set_facecolor("white")

    # Helper to draw one set of lines
    def draw_lines(step, lw, col):
        # Vertical lines, both sides of 0
        for x in np.arange(0, xmax + step, step):
            ax.axvline(x, linewidth=lw, color=col)
        for x in np.arange(0, xmin - step, -step):
            ax.axvline(x, linewidth=lw, color=col)

        # Horizontal lines, both sides of 0
        for y in np.arange(0, ymax + step, step):
            ax.axhline(y, linewidth=lw, color=col)
        for y in np.arange(0, ymin - step, -step):
            ax.axhline(y, linewidth=lw, color=col)

    # Draw in order: minor → medium → major
    draw_lines(minor_step, 1.0, "lightgrey")
    draw_lines(medium_step, 2.0, "grey")
    draw_lines(major_step, 4.0, "black")

    # Remove ticks
    ax.tick_params(which="both", bottom=False, left=False,
                   labelbottom=False, labelleft=False)

    fig.tight_layout(pad=0)

    if filename:
        fig.savefig(filename, dpi=dpi, bbox_inches="tight", pad_inches=0)
        print(f"Saved grid to {filename}")

    # Convert to numpy RGBA [0–1]
    fig.canvas.draw()
    buf = np.frombuffer(fig.canvas.tostring_argb(), dtype=np.uint8)
    buf = buf.reshape(height_px, width_px, 4)
    # ARGB → RGBA
    buf = buf[:, :, [1, 2, 3, 0]]
    arr = buf.astype(np.float32) / 255.0

    return arr



from modules.module import Module, debug_decorator
import numpy as np
from utils.terrains import Terrain


class Generative(Module):
    create_folder = False
    '''
    Generative base
    '''
    @debug_decorator
    def __call__(
            self, function_name=None, terrain_temp=None, default=None,
            position=[[0, 0]], height=[], width=[],
            aspect=[], yaw_deg=[], pitch_deg=[],
            size=None, grid_size=None, resolution=None, extent=None,
            reset_input=True,
            **kwargs):
        '''

        Args:
          position: list of positions, where position is [x, y] list
          heights: list of heights, where height is float in [-infty, infty]
          yaws: list of yaw, where yaw is float in [0, 2*np.pi]
        '''
        terrain_temp = [] if terrain_temp is None else terrain_temp
        functions = ['gaussian', 'step', 'donut', 'plane', 'sphere', 'cube',
                     'smoothstep', 'sine']
        print(f"function_name:{function_name}")
        if function_name is None:
            function_name = np.random.choice(functions)

        assert function_name in functions

        # Format arguments
        if np.array(position).ndim == 1:
            position = [position]
        if np.array(height).ndim == 0:
            height = [height]
        if np.array(yaw_deg).ndim == 0:
            yaw_deg = [yaw_deg]
        if np.array(width).ndim == 0:
            width = [width]
        if np.array(aspect).ndim == 0:
            aspect = [aspect]
        if np.array(pitch_deg).ndim == 0:
            pitch_deg = [pitch_deg]

        # Setup sizes etc.
        from utils.artificial_shapes import determine_extent_and_resolution
        # Get size and resolution from any two tuples: resolution, size, resolution
        extent, (N_x, N_y) = determine_extent_and_resolution(
            resolution, size, grid_size, extent)

        properties = {
            'position': position,
            'height': height,
            'yaw_deg': yaw_deg,
            'width': width,
            'aspect': aspect,
            'pitch_deg': pitch_deg,
        }

        keys = list(properties.keys())

        from itertools import zip_longest
        from utils.artificial_shapes import get_meshgrid
        from utils.artificial_shapes import FUNCTIONS

        for values in zip_longest(*properties.values()):
            # Create a dictionary for each group, excluding any None values
            kwargs = {k: v for k, v in zip(keys, values) if v is not None}
            self.logger.debug(f"kwargs:{kwargs}")

            # Setup function callable
            function_callable = FUNCTIONS[function_name](**kwargs)

            # Get meshgrid
            X, Y = get_meshgrid(extent, N_x, N_y)

            # Get heights from function callable
            heights_array = function_callable(X, Y)

            # Make terrain, and add to dict
            terrain = Terrain.from_array(heights_array, extent=extent)
            # desctiption = f'{kwargs.__repr__()}'
            desctiption = f'{self.name}{self.file_id}_{kwargs.__repr__()}'
            terrain_temp.append(terrain)

        pipe = {
            'terrain_temp': terrain_temp,
        }

        if reset_input:
            pipe['position'] = [[0, 0]]
            for parameter in ['height', 'yaw_deg',
                              'width', 'aspect', 'pitch_deg']:
                if parameter:
                    pipe[parameter] = []

        return pipe


class Gaussian(Generative):
    def __call__(self, *args, **kwargs):
        return super().__call__(
            *args, function_name='gaussian', **kwargs)


class Step(Generative):
    def __call__(self, *args, **kwargs):
        return super().__call__(
            *args, function_name='step', **kwargs)


class Donut(Generative):
    def __call__(self, *args, **kwargs):
        return super().__call__(
            *args, function_name='donut', **kwargs)


class Plane(Generative):
    def __call__(self, *args, **kwargs):
        return super().__call__(
            *args, function_name='plane', **kwargs)


class Sphere(Generative):
    def __call__(self, *args, **kwargs):
        return super().__call__(
            *args, function_name='sphere', **kwargs)


class Cube(Generative):
    def __call__(self, *args, **kwargs):
        return super().__call__(
            *args, function_name='cube', **kwargs)


class SmoothStep(Generative):
    def __call__(self, *args, **kwargs):
        return super().__call__(
            *args, function_name='smoothstep', **kwargs)


class Sine(Generative):
    def __call__(self, call_number=None, call_total=None, *args, **kwargs):
        return super().__call__(
            *args, function_name='sine', call_number=call_number,
            call_total=call_total, **kwargs)


class Function(Module):
    create_folder = False
    '''
    Execute given x, y function
    '''
    @debug_decorator
    def __call__(
            self,
            expression='x',
            terrain_temp=None, default=None, overwrite=None,
            size=None, grid_size=None, resolution=None, extent=None,
            **kwargs):
        '''
        '''
        expression = default if default is not None else expression

        terrain_temp = [] if terrain_temp is None else terrain_temp

        # Setup sizes etc.
        from utils.artificial_shapes import determine_extent_and_resolution
        # Get size and resolution from any two tuples: resolution, size, resolution
        extent, (N_x, N_y) = determine_extent_and_resolution(
            resolution, size, grid_size, extent)

        from utils.artificial_shapes import get_meshgrid

        # Get meshgrid
        X, Y = get_meshgrid(extent, N_x, N_y)

        safe_locals = {
            'x': X,
            'y': Y,
            'r': np.sqrt(np.square(X) + np.square(Y)),
            'np': np  # Optional: allow np.sin, np.sqrt, etc.
        }

        # Evaluate the expression
        heights_array = eval(expression, {"__builtins__": {}}, safe_locals)

        terrain = Terrain.from_array(heights_array, extent=extent)
        # desctiption = f'{kwargs.__repr__()}'
        desctiption = f'{self.name}{self.file_id}_{kwargs.__repr__()}'
        terrain_temp.append(terrain)

        pipe = {
            'terrain_temp': terrain_temp,
        }

        return pipe



# Change name of this file!
import numpy as np
import sys
import inspect


def get_meshgrid(extent, N_x, N_y):
    # Generating the x and y coordinates
    x = np.linspace(extent[0], extent[1], N_x, endpoint=False)
    y = np.linspace(extent[2], extent[3], N_y, endpoint=False)
    X, Y = np.meshgrid(x, y, indexing='ij')

    return X, Y


def rotate_meshgrid(X, Y, theta_degree):
    '''

    Args:
      theta_degree: angle, in degrees
    '''
    theta = theta_degree/180.0*np.pi

    # Create rotation matrix
    rotation_matrix = np.array([
        [np.cos(theta), -np.sin(theta)],
        [np.sin(theta), np.cos(theta)]
    ])

    # Flatten the grids
    x_flat = X.flatten()
    y_flat = Y.flatten()

    # Apply the rotation to each point
    # Stack x and y for matrix multiplication
    xy_points = np.vstack([x_flat, y_flat])
    # Matrix multiplication
    rotated_points = rotation_matrix @ xy_points

    # Reshape the rotated points back to grid shape
    X_rotated = rotated_points[0, :].reshape(X.shape)
    Y_rotated = rotated_points[1, :].reshape(Y.shape)

    return X_rotated, Y_rotated


def cartesian_to_polar(x, y):
    rho = np.sqrt(x**2 + y**2)
    phi = np.arctan2(y, x)

    return (rho, phi)


def determine_extent_and_resolution(resolution=None, size=None, grid_size=None, extent=None):
    '''
    Determine extent and resolution.

    Default: use extent (m) and resolution (points/m) to decide -> (extent, grid_size)
    
    If grid_size explicitly given, use that instead.
    '''
    if resolution is not None:
        # Default, use resolution
        if isinstance(resolution, list):
            resolution_x, resolution_y = resolution
        else:
            resolution_x, resolution_y = (resolution, resolution)

        from utils.terrains import extent_to_size
        [size_x, size_y] = extent_to_size(extent)
        grid_size_x = int(size_x * resolution_x)
        grid_size_y = int(size_y * resolution_y)
    elif grid_size is not None:
        # But use grid-size if explicitly given
        if isinstance(grid_size, list):
            grid_size_x, grid_size_y = grid_size
        else:
            grid_size_x, grid_size_y = (grid_size, grid_size)

    return extent, [grid_size_x, grid_size_y]


def gaussian_1d(x, mu=0, sigma=1):
    return np.exp(-0.5 * ((x - mu) / sigma)**2)


def smoothstep(x):
    # 'smoothstep' between x=0 and x=1, using 3x^2 - 2x^3
    return np.where(x <= 0, 0, np.where(x >= 1, 1, 3 * x**2 - 2 * x**3))


class GaussianFunction():
    def __init__(self, position=(0, 0), height=2.5, width=5, yaw_deg=0, aspect=1,
                 **_):
        self.pos = position
        self.r_x = 0.5*width / np.sqrt(aspect)
        self.r_y = 0.5*width * np.sqrt(aspect)
        self.A = height
        self.yaw_deg = yaw_deg

    def __call__(self, x, y):
        # Adjust position
        # NOTE: 'shifted, maybe due to our 'transpose' habit
        x = x - self.pos[0]
        y = y - self.pos[1]

        # Rotate
        x, y = rotate_meshgrid(x, y, self.yaw_deg)

        return self.A * gaussian_1d(x, sigma=self.r_x)\
            * gaussian_1d(y, sigma=self.r_y)


class SphereFunction():
    def __init__(self, position=(0, 0), height=2.5, width=5, yaw_deg=0, aspect=1,
                 **_):
        '''
          width: diameter of sphere
          height: maximum height
          aspect:
        '''
        self.pos = position
        self.height = height
        self.yaw_deg = yaw_deg
        self.radius = width/2
        self.aspect = aspect

    def __call__(self, x, y):
        # Adjust position
        # NOTE: 'shifted, maybe due to our 'transpose' habit
        x = x - self.pos[0]
        y = y - self.pos[1]

        # Rotate
        x, y = rotate_meshgrid(x, y, self.yaw_deg)

        # Calculate distance from center after rotation
        distance = np.sqrt((x*np.sqrt(self.aspect))**2 + (y/np.sqrt(self.aspect))**2)

        # Calculate height
        sphere_height = np.sqrt(self.radius**2 - distance**2)
        # Adjust sphere position sphere in z
        sphere_height = sphere_height - self.radius + self.height
        # Set to 0 outside
        sphere_height = np.where(sphere_height >= 0, sphere_height, 0)

        return sphere_height


class StepFunction():
    def __init__(self, position=(0, 0), height=2.5, yaw_deg=0, **_):
        self.pos = position
        self.A = height
        self.yaw_deg = yaw_deg

    def __call__(self, x, y):
        # Adjust position
        # NOTE: 'shifted, maybe due to our 'transpose' habit
        x = x - self.pos[0]
        y = y - self.pos[1]

        x, y = rotate_meshgrid(x, y, self.yaw_deg)

        return self.A * np.greater(x, 0)


class SmoothStepFunction():
    def __init__(self, position=(0, 0), height=2.5, yaw_deg=0, **_):
        self.pos = position
        self.A = height
        self.yaw_deg = yaw_deg
        self.step_length = height

    def __call__(self, x, y):
        # Adjust position
        # NOTE: 'shifted, maybe due to our 'transpose' habit
        x = x - self.pos[0]
        y = y - self.pos[1]

        # Rotate
        x, y = rotate_meshgrid(x, y, self.yaw_deg)

        # Calculate smoothstep function for each element in x
        smooth_x = smoothstep((x) / self.step_length)

        return self.A * smooth_x


class DonutFunction():
    # Even if this a Gaussian shape
    def __init__(self, position=(0, 0), height=2.5, width=5, yaw_deg=0, aspect=1,
                 **_):
        self.pos = position
        self.A = height
        self.radius = width/2
        self.aspect = aspect
        self.yaw_deg = yaw_deg

    def __call__(self, x, y):
        # Adjust position
        # NOTE: 'shifted, maybe due to our 'transpose' habit
        x = x - self.pos[0]
        y = y - self.pos[1]

        # Rotate
        x, y = rotate_meshgrid(x, y, self.yaw_deg)

        # Stretch coordinates
        x = x * np.sqrt(self.aspect)
        y = y / np.sqrt(self.aspect)

        # Turn to polar coordinates
        rho, phi = cartesian_to_polar(x, y)

        return self.A * gaussian_1d(rho, mu=self.radius, sigma=self.A/2)


class PlaneFunction():
    def __init__(self, position=(0, 0), pitch_deg=0, yaw_deg=0, **_):
        self.pitch_deg = pitch_deg
        self.yaw_deg = yaw_deg
        self.pos = position
        # TODO: Fixa så att planet går igenom punkten!

    def __call__(self, x, y):
        # Rotate
        x, y = rotate_meshgrid(x, y, self.yaw_deg)

        # Setup angle
        tan = np.tan(self.pitch_deg*np.pi/180.0)

        return y * tan


class SineFunction():
    def __init__(self, yaw_deg=0, height=1, width=2, **_):
        self.yaw_deg = yaw_deg
        self.amplitude = height/2
        self.width = width

    def __call__(self, x, y):
        # Rotate
        x, y = rotate_meshgrid(x, y, self.yaw_deg)

        return self.amplitude * np.sin(y*2*np.pi/self.width)


class CubeFunction():
    def __init__(self, position=(0, 0), height=2.5, width=5, yaw_deg=0, aspect=1, **_):
        self.position = position
        self.height = height
        self.width_x = width * np.sqrt(aspect)
        self.width_y = width / np.sqrt(aspect)
        self.yaw_deg = yaw_deg
        self.aspect = aspect

    def __call__(self, x, y):
        # Adjust position
        # NOTE: 'shifted, maybe due to our 'transpose' habit
        x = x - self.position[0]
        y = y - self.position[1]

        # Rotate
        x, y = rotate_meshgrid(x, y, self.yaw_deg)

        # Calculate cube
        x_dir = np.logical_and(x > -self.width_x/2, x < self.width_x/2)
        y_dir = np.logical_and(y > -self.width_y/2, y < self.width_y/2)

        return np.logical_and(x_dir, y_dir) * self.height


clsmembers_pairs = inspect.getmembers(sys.modules[__name__], inspect.isclass)
FUNCTIONS = {k.replace('Function', '').lower(): v for (k, v)
             in clsmembers_pairs if 'Function' in k}



import argparse
import numpy as np
import os
from utils.terrains import Terrain
from itertools import cycle, islice


def calc_y(array, ha=1.0):
    shape = array.shape[:-1]
    # Reshape to two dimensions
    array = array.reshape(-1, 4)

    # Rescale classes given area
    # rikligt = np.array([4000, 40000])*ha
    måttligt = np.array([400, 4000])*ha
    sparsamt = np.array([40, 400])*ha
    enstaka = np.array([4, 40])*ha

    # if H[20] < sparsamt[1]:
    a = array[:, 0] < sparsamt[1]
    # if H[40] + H[60] + H[80] < enstaka[1]:
    b = array[:, 1] + array[:, 2] + array[:, 3] < enstaka[1]
    # elif H[40] < sparsamt[1] and H[60] + H[80] < enstaka[1]:
    c1 = array[:, 1] < sparsamt[1]
    c2 = array[:, 2] + array[:, 3] < enstaka[1]
    c = np.logical_and(c1, c2)
    c = np.logical_and(c, np.logical_not(b))
    # elif H[40] < måttligt[1] and H[60] < sparsamt[1]:  # d
    d1 = array[:, 1] < måttligt[1]
    d2 = array[:, 2] < sparsamt[1]
    d = np.logical_and(d1, d2)
    d = np.logical_and(d, np.logical_not(np.logical_or(b, c)))
    # if H[80] < enstaka[1]:
    e = array[:, 3] < enstaka[1]
    # elif H[80] < sparsamt[1]:
    f1 = enstaka[1] < array[:, 3]
    f2 = array[:, 3] < sparsamt[1]
    f = np.logical_and(f1, f2)
    # not e or f
    g = np.logical_not(np.logical_or(e, f))
    # not b, d, or d
    h = np.logical_not(b | c | d)
    # elif sparsamt[1] < H[20] < måttligt[1]: # i
    i1 = sparsamt[1] < array[:, 0]
    i2 = array[:, 0] < måttligt[1]
    i = np.logical_and(i1, i2)
    # if H[40] + H[60] + H[80] < enstaka[0]: # j
    j = array[:, 1] + array[:, 2] + array[:, 3] < enstaka[0]
    # not a or i
    k = np.logical_not(np.logical_or(a, i))

    y1_1 = np.logical_and(a, b)
    y2_1 = np.logical_and(a, c)
    y3_1 = a & d & e
    y4_1 = a & d & f
    y5_1 = a & d & g
    y5_2 = np.logical_and(a, h)

    y1_2 = np.logical_and(i, j)
    y2_2 = np.logical_and(i, c)
    y3_2 = i & d & e
    y4_2 = i & d & f
    y5_3 = i & d & g
    y5_4 = np.logical_and(i, h)

    y2_3 = np.logical_and(k, j)
    y3_3 = k & d & e
    y4_3 = k & d & f
    y5_5 = k & d & g
    y5_6 = np.logical_and(k, h)

    # Combine using logical_or
    y1 = y1_1 | y1_2
    y2 = y2_1 | y2_2 | y2_3
    y3 = y3_1 | y3_2 | y3_3
    y4 = y4_1 | y4_2 | y4_3
    y5 = y5_1 | y5_2 | y5_3 | y5_4 | y5_5 | y5_6

    # Construct y integer
    y = y1*1 + y2*2 + y3*3 + y4*4 + y5*5

    return y.reshape(shape)


def calc_ytstruktur(H_to_number_mapping, ha=1.0):
    # Take input as either list of dict
    is_array = isinstance(H_to_number_mapping, np.ndarray)
    is_list = isinstance(H_to_number_mapping, list)
    if is_list or is_array:
        H = {
            20: H_to_number_mapping[0],
            40: H_to_number_mapping[1],
            60: H_to_number_mapping[2],
            80: H_to_number_mapping[3],
        }
    else:
        H = H_to_number_mapping

    Y = 0
    rikligt = np.array([4000, 40000])*ha
    måttligt = np.array([400, 4000])*ha
    sparsamt = np.array([40, 400])*ha
    enstaka = np.array([4, 40])*ha

    # H20 sparsamt
    if H[20] < sparsamt[1]: # a
        if H[40] + H[60] + H[80] < enstaka[1]: # b
            Y = 1
        elif H[40] < sparsamt[1] and H[60] + H[80] < enstaka[1]: # c
            Y = 2
        elif H[40] < måttligt[1] and H[60] < sparsamt[1]:  # d
            if H[80] < enstaka[1]: # e
                Y = 3
            elif H[80] < sparsamt[1]: # f
                Y = 4
            else: # g
                Y = 5
        else: # h
            Y = 5
    # H20 måttligt
    elif sparsamt[1] < H[20] < måttligt[1]: # i
        # H40-H80 finns inga
        if H[40] + H[60] + H[80] < enstaka[0]: # j
            Y = 1
        # H40 sparsamt, H60-H80 enstaka
        elif H[40] < sparsamt[1] and H[60] + H[80] < enstaka[1]: # c
            Y = 2
        # H40 måttligt
        elif H[40] < måttligt[1] and H[60] < sparsamt[1]: # d
            if H[80] < enstaka[1]: # e
                Y = 3
            elif H[80] < sparsamt[1]: # f
                Y = 4
            else:
                Y = 5 # g
        else: # h
            Y = 5
    # H20 Rikligt
    else: # k
        if H[40] + H[60] + H[80] < enstaka[0]: # j
            Y = 2
        # H40 måttligt
        elif H[40] < måttligt[1] and H[60] < sparsamt[1]: # d
            if H[80] < enstaka[1]: # e
                Y = 3
            elif H[80] < sparsamt[1]: # f
                Y = 4
            else: # g
                Y = 5
        else: # h
            Y = 5

    return Y


def calc_ytstruktur_2(h_20, h_40, h_60, h_80):
    '''
    Def rocks per hectare in each size class
    '''
    def classify(v):
        ''' Classify according to enstaka, sparsamt, måttlit, rikligt '''
        return 0 if v < 4 else 1 if v < 40 else 2 if v < 400 else 3 if v < 4000 else 4

    number_mapping = {
        0: 'Finns inga',
        1: 'Enstaka',
        2: 'Sparsamt',
        3: 'Måttligt',
        4: 'Rikligt',
    }
    none = 0
    occasional = 1
    sparse = 2
    moderate = 3
    abundant = 4

    # Calculate
    if classify(h_20) <= sparse:
        if classify(h_40 + h_60 + h_80) <= occasional:
            Y = 1
        elif classify(h_40) <= sparse and classify(h_60 + h_80) <= occasional:
            Y = 2
        elif classify(h_40) <= moderate:
            if classify(h_60) <= sparse:
                if classify(h_80) <= occasional:
                    Y = 3
                elif classify(h_80) <= sparse:
                    Y = 4
                else:
                    Y = 5
            else:
                Y = 5
        else:
            Y = 5

    elif classify(h_20) == moderate:
        if classify(h_40 + h_60 + h_80) == none:
            Y = 1
        elif classify(h_40) <= sparse and classify(h_60 + h_80) <= occasional:
            Y = 2
        elif classify(h_40) <= moderate:
            if classify(h_60) <= sparse:
                if classify(h_80) <= occasional:
                    Y = 3
                elif classify(h_80) <= sparse:
                    Y = 4
                else:
                    Y = 5
            else:
                Y = 5
        else:
            Y = 5

    else:  # classify(h_20) == abundant
        if classify(h_40 + h_60 + h_80) == none:
            Y = 2
        elif classify(h_40) <= moderate:
            if classify(h_60) <= sparse:
                if classify(h_80) <= occasional:
                    Y = 3
                elif classify(h_80) <= sparse:
                    Y = 4
                else:
                    Y = 5
            else:
                Y = 5
        else:
            Y = 5

    return Y


def generate_size_distribution(Y=3, random=False):
    '''
    Generate a size distribution
    '''
    rikligt = [4000, 40000]
    måttligt = [400, 4000]
    sparsamt = [40, 400]
    enstaka = [4, 40]

    if Y == 3 and not random:
        H_to_number_mapping = {
            20: 8000,
            40: 2000,
            60: 200,
            80: 20,
        }
    elif Y == 3 and random:
        H_to_number_mapping = {
            20: np.random.randint(*rikligt),
            40: np.random.randint(*måttligt),
            60: np.random.randint(*sparsamt),
            80: np.random.randint(*enstaka),
        }

    return H_to_number_mapping


class Grid():
    def __init__(self, shape=None, size=None, cell_size=None):
        if size is None and cell_size is None and shape is None:
            # Default
            size = (100, 100)
            shape = (32, 32)
        elif size is None and cell_size is None:
            size = (100, 100)
        elif size is None and shape is None:
            shape = (32, 32)
            size = np.multiply(shape, cell_size)
        elif cell_size is None and shape is None:
            shape = (32, 32)
        elif size is None:
            size = np.multiply(shape, cell_size)
        elif shape is None:
            shape = np.divide(size, cell_size).astype(np.uint64).tolist()
        elif cell_size is None:
            pass
        else:
            raise Exception("Grid: Can't specify size, cell_size, and shape")

        # Define coordinates, representing edges of bins
        x_edges = np.linspace(0, size[0], shape[0]+1)
        y_edges = np.linspace(0, size[1], shape[1]+1)

        x_size = (x_edges[1]-x_edges[0])
        y_size = (y_edges[1]-y_edges[0])

        x_centers = x_edges[:-1] + x_size/2
        y_centers = y_edges[:-1] + y_size/2

        self.shape = shape
        self.size = size

        self.x_centers = x_centers
        self.y_centers = y_centers
        self.x_size = x_size
        self.y_size = y_size
        self.x_edges = x_edges
        self.y_edges = y_edges

    def get_points(self):
        for x in range(self.shape[0]):
            for y in range(self.shape[1]):
                yield np.array((self.x_centers[x], self.y_centers[y]))

    @property
    def area(self):
        return self.x_size * self.y_size

    @property
    def area_ha(self):
        return self.area/1e+4

    @property
    def extent(self):
        x_edges = self.x_edges
        y_edges = self.y_edges
        return [x_edges[0], x_edges[-1], y_edges[0], y_edges[-1]]

    def __str__(self):
        return f"Grid: shape={self.shape}, area={self.area_ha:.2f} ha"

    def __repr__(self):
        return f"Grid(size={self.size}, shape={self.shape})"


def count_obstacles(
        grid, obstacles,
        only_center=False,
        return_particle_count=False
):
    '''
    Count number of obstacles, in every grid cell

    Return:
      counts array, of shape (*grid.shape, 4)
        where the last dim is a 4-list e.g. [100, 34, 10, 1]
        (i.e. the number of H20, H40, H60 H80)
    '''
    position = obstacles.position
    radius = obstacles.radius

    list_of_counts = []
    accumulated_particle_count = np.zeros_like(radius)

    for x, y in grid.get_points():
        # Pick obstacles within the grid box
        if only_center:
            x_pick = (np.absolute(x - position[0, :]) < grid.x_size/2)
            y_pick = (np.absolute(y - position[1, :]) < grid.y_size/2)
        else:
            x_pick = (np.absolute(x - position[0, :]) - radius[0, :] < grid.x_size/2)
            y_pick = (np.absolute(y - position[1, :]) - radius[0, :] < grid.y_size/2)
        pick = np.logical_and(x_pick, y_pick).squeeze()
        accumulated_particle_count += pick
        diameter = 2*radius[:, pick].flatten()
        diameter = np.clip(diameter, 0, 0.9)

        # Divide into boxes
        bins = [0.10, 0.30, 0.50, 0.70, 0.90]
        counts, bin_edges = np.histogram(diameter, bins=bins)
        list_of_counts.append(counts)

    count_array = np.array(list_of_counts).reshape(*grid.shape, 4)

    if return_particle_count:
        return count_array, accumulated_particle_count
    else:
        return count_array


def check_double_counts(grid, obstacles):
    '''
    Construct array of how many cells an obstacle is in
    '''
    radius = obstacles.radius[0, :]
    pos_x = obstacles.position[0, :]
    pos_y = obstacles.position[1, :]

    # Get obstacles which are on one or several edges
    x_pick = np.mod(pos_x+radius, grid.x_size) < 2*radius
    y_pick = np.mod(pos_y+radius, grid.y_size) < 2*radius

    pick_corners = np.logical_and(x_pick, y_pick)
    pick_edges = np.logical_xor(x_pick, y_pick)
    pick_corners_and_edges = np.logical_or(x_pick, y_pick)
    pick_neither = np.logical_not(pick_corners_and_edges)

    return (pick_corners * 4 + pick_edges * 2 + pick_neither).reshape(1, -1)


def spawn_particles(obstacle_density_mapping, probability):
    '''
    Generate particles arrays from a 2d probability (for location)
    and a mapping from obstacle size (H) to number of obstacles

    Returns:
      dict of arrays
      particle_pos has size [2, N]
      particle_radius has size [1, N]
      {
    '''
    from utils.noise import generate_points_from_2D_prob
    # Create lists of particles
    particle_pos = []
    particle_width = []
    for H, density in obstacle_density_mapping.items():
        # Generate random integers
        size = probability.shape
        number = round(density * size[0] * size[1]/100**2)
        indices = generate_points_from_2D_prob(probability, N=number)
        # Add uniform noise to turn into float
        rng = np.random.default_rng()
        uniform_noise = rng.uniform(size=indices.shape)
        position = indices + uniform_noise

        particle_pos.append(position)
        particle_width.append(H/100.0 * np.ones((1, indices.shape[1])))

    # Turn into arrays
    particle_pos = np.concatenate(particle_pos, axis=1)
    particle_width = np.concatenate(particle_width, axis=1)

    # Generate other parameters
    N = particle_pos.shape[1]
    yaw_deg = np.random.uniform(low=0, high=360, size=(1, N))
    pitch_deg = np.random.uniform(low=0, high=30, size=(1, N))
    aspect = np.random.uniform(low=0.5, high=1.5, size=(1, N))

    return {
        'position': particle_pos,
        'width': particle_width,
        'height': particle_width/2,
        'yaw_deg': yaw_deg,
        'aspect': aspect,
        'pitch_deg': pitch_deg,
        }


def get_list_of_files(dirs: list, pattern: str):
    '''
    Return a (sorted) list of all files (with directories) in the
    directories dirs
    '''
    out = []
    for d in dirs:
        files = sorted(os.listdir(d))
        j = 0
        for f in files:
            # print(f"f:{f}")
            if pattern in f:
                j += 1
                out.append(f)

    return sorted(out)


def add_id(filename, file_id):
    '''
    Add file_id (e.g. '_00005' or '') to filename
    '''
    root, extension = os.path.splitext(filename)
    filename = root + file_id + extension

    return filename


def get_terrains(
        terrain_temp=[], terrain_prim=[],
        last=None, print_fn=print, remove=True):
    ''' Get terrains from temp/heap

    Terrains are removed from the temp/heap in place if `remove` is True.
    '''

    if last is not None and last > 0:
        # When e.g. taking the last two, the slice is [-2:]
        # So here we make sure this is negative
        last = -last

    if len(terrain_temp) > 0:
        # Get terrains from temp (if last=None then all are used)
        terrains = terrain_temp[last:]
        print_fn(f"Use {len(terrains)}/{len(terrain_temp)} terrains from temp")
        # Remove any used terrains from terrain_temp in place
        if remove:
            del terrain_temp[last:]

    elif len(terrain_prim) > 0:
        # Get terrains from heap (if last=None then all are used)
        terrains = terrain_prim[last:]
        print_fn(f"Use {len(terrains)}/{len(terrain_prim)} terrains from primary")

        # Remove any used terrains from terrain_prim in place
        if remove:
            del terrain_prim[last:]
    else:
        raise AttributeError("Both terrain_temp and terrain_prim are empty")

    return terrains


def get_terrain(*args, last=None, **kwargs):
    return get_terrains(*args, last=-1, **kwargs)[0]


def find_rocks(terrain):
    """
    Identify connected high regions in a terrain and return their real-world
    centroids, maximum heights, and pixel counts.

    Returns:
    - positions: Nx2 array of (x, y) centroids in real-world coordinates
    - heights: Maximum elevation per region
    - sizes: Number of pixels in each region
    """
    import numpy as np
    import scipy.ndimage as ndimage

    labeled, num = ndimage.label(terrain.array)
    positions = []
    heights = []
    sizes = []

    for label in range(1, num + 1):
        coords = np.argwhere(labeled == label)
        sizes.append(len(coords))

        center = np.mean(coords, axis=0)
        max_idx = np.argmax(terrain.array[labeled == label])
        max_pos = coords[max_idx]
        max_val = terrain.array[max_pos[0], max_pos[1]]

        # Convert row/col center to (x, y) coordinates
        x_range = [0, terrain.array.shape[0]]
        y_range = [0, terrain.array.shape[1]]

        x = np.interp(center[0], x_range, terrain.extent[:2])
        y = np.interp(center[1], y_range, terrain.extent[2:4])
        positions.append([x, y])
        heights.append(max_val)

    return np.array(positions), np.array(heights), np.array(sizes)


class Distribution:
    def __init__(self, dist_name, *params):
        self.dist_name = dist_name
        self.params = params
        self._sequence = None
        self._sequence_iter = None
        self.dist = self._create_distribution()

    def _create_distribution(self):
        # Random distributions from numpy
        random_distributions = {
            'uniform': np.random.uniform,
            'normal': np.random.normal,
            'exponential': np.random.exponential,
            'beta': np.random.beta,
        }

        regular_distributions = {
            'arange': np.arange,
            'linspace': np.linspace,
            'logspace': np.logspace,
        }

        if self.dist_name in random_distributions:
            return lambda size: random_distributions[self.dist_name](
                *self.params, size=size)
        elif self.dist_name in regular_distributions:
            self._sequence = regular_distributions[self.dist_name](*self.params)
            self._sequence_iter = cycle(self._sequence)
            return self._sample_from_sequence
        else:
            raise ValueError(f"Unsupported distribution: {self.dist_name}")

    def _sample_from_sequence(self, size=1):
        return np.array(list(islice(self._sequence_iter, size)))

    def sample(self, size=1):
        return self.dist(size=size)

    def __call__(self, *args, **kwargs):
        return self.sample(*args, **kwargs)

    def __repr__(self):
        return f"<Distribution: {self.dist_name} params={self.params}>"


def parse_and_assign_distribution(expression):
    import re
    # Replace square parenthesis with soft
    expression = expression.replace('[', '(').replace(']', ')')
    # logger.debug(f"expression:{expression}")
    match = re.match(r"(\w+)=([\w_]+)\((.*)\)", expression)
    # match = re.match(r"(\w+):(\w+)=([\w_]+)\((.*)\)", expression)
    if not match:
        raise ValueError(f"Invalid expression format: {expression}")

    var_name, dist_name, params_str = match.groups()
    params = eval(f"({params_str})")

    # Make sure params is a tuple
    if not isinstance(params, tuple):
        params = (params,)

    dist_obj = Distribution(dist_name, *params)
    return var_name, dist_obj



import os
import sys
try:
    import bpy
    # Make sure current dir is in path. A blender py workaround
    dir = os.path.dirname(bpy.data.filepath)
    if dir not in sys.path:
        sys.path.append(dir)
except Exception:
    using_blender = False
else:
    from modules.blender import fix_blender_argv, fix_blender_path
    fix_blender_path()
    fix_blender_argv()
    using_blender = True

from utils.parse_utils import create_parser, get_args_combined_with_settings
from modules.modules import MODULES
from utils.logging_utils import get_logger, level_map


def main():
    parser = create_parser(specific_args=['modules'])
    args = get_args_combined_with_settings(parser)

    # Create folder if needed
    save_dir = args.save_dir
    if save_dir and not os.path.exists(save_dir):
        os.makedirs(save_dir)

    # Add logger and set level
    logger = get_logger(save_dir=save_dir)
    logger.setLevel(level_map[args.debug])

    # Parse settings
    settings = args.settings if args.settings is not None else {}

    # Check that all modules exist
    for key, value in args.modules:
        if key not in MODULES.keys():
            logger.info(f"Invalid choice {key}. ({MODULES.keys()})")
            exit(1)

    if using_blender:
        # Prepend the 'BasicSetup' and 'Holdout' modules
        args.modules = [('BasicSetup', None), ('Holdout', None)] + args.modules

    pipe = {
        'size': 50,  # default
        'resolution': 2,  # default
        'extent': [-25, 25, -25, 25],  # default
        'call_number': 0,
        'call_total': 1,
    }

    # Extract general settings (not given by dicts)
    general_kwargs = {
    }
    for k, v in settings.items():
        if type(v) is not dict:
            general_kwargs[k] = v

    list_of_modules_kwargs_tuples = []

    # Setup list of modules and kwargs
    for module, options in args.modules:
        # Skip if no modules
        if module is None:
            continue
        module_class = MODULES[module]
        module_obj = module_class(save_dir, logger)
        # Extract possible settings kwargs
        if isinstance(options, dict):
            kwargs = options
        elif options is None:
            kwargs = {}
        else:
            # Store value as 'default'
            kwargs = {'default': options}

        # combine with general kwargs
        kwargs = {**general_kwargs, **kwargs}
        # Add to list
        list_of_modules_kwargs_tuples.append((module_obj, kwargs))

    # Start recursive call of modules
    recursive_module_call(
        list_of_modules_kwargs_tuples, pipe=pipe, logger=logger,
    )


def recursive_module_call(
        list_of_modules_kwargs_tuples, index=0, pipe={}, logger=None):
    '''
    Call next level recursivly
    '''
    logger.debug(f"### index: {index}")
    # Base case: reached the end of the module list
    if index >= len(list_of_modules_kwargs_tuples):
        return pipe

    # Get module and kwargs
    module_obj, kwargs = list_of_modules_kwargs_tuples[index]
    logger.debug(f"  module_obj: {module_obj}")
    logger.debug(f"  kwargs: {kwargs}")
    logger.debug(f"  pipe: {pipe}")

    # Test: kwargs overrides pipe in input?
    module_obj.start(**{**pipe, **kwargs})
    for returned_data in module_obj:
        if isinstance(returned_data, dict):
            pipe = {**pipe, **returned_data}

        pipe = recursive_module_call(
            list_of_modules_kwargs_tuples, index+1, pipe, logger)
        # # Update pipe, with data from the module to 'the right'
        # if isinstance(return_from_right, dict):
        #     pipe = {**pipe, **return_from_right}

    # # When we return something here, it will be given to the module on the left
    return pipe


if __name__ == "__main__":
    main()



from modules.module import Module
import numpy as np
from itertools import zip_longest



class Loop(Module):
    ''' Basic loop '''
    create_folder = False

    def start(self, default=None, n_loops=2, loop_id=None, loop_id_r=None, call_total=None,
              call_number=None, parameter=None, expression=None, values=None,
              **kwargs):

        # Handle if default is given a parameter=expression
        if isinstance(default, str) and '=' in default:
            try:
                # First try to handle like a distribution
                # Note: only works for 'regular' distributions, not random ones
                # There needs to be a fixed length
                # e.g.
                from utils.utils import parse_and_assign_distribution
                parameter, distribution = parse_and_assign_distribution(default)
                # Get the sequence from the (regular) distribution
                sequence = distribution._sequence
                self.n_loops = len(sequence)
                values = sequence

            except ValueError:
                parameter, expression = default.split('=')
                values = eval(expression)
                self.n_loops = len(values)

        elif isinstance(default, int):
            # Length of the loop
            self.n_loops = default if default is not None else n_loops
        elif isinstance(default, str) and 'x' in default:
            try:
                a, b = map(int, default.lower().split('x'))
                self.n_loops = a * b
            except ValueError:
                raise ValueError(f"Invalid format for default: {default}. Expected format like '2x3'.")
            else:
                # Construct a list of extent 'edges'
                x_min, x_max, y_min, y_max = kwargs['extent']
                x_split = np.linspace(x_min, x_max, a+1).astype(float).tolist()
                y_split = np.linspace(y_min, y_max, b+1).astype(float).tolist()
                # Construct a list of new extents
                self.list_of_new_extents = []
                for i in range(a):
                    for j in range(b):
                        extent = x_split[i:i+2] + y_split[j:j+2]
                        self.list_of_new_extents.append(extent)
        else:
            pass

        # Store any parameter or values
        self.parameter = parameter
        self.values = values if values is not None else []

        # Create a instance of the generator
        self.loop_generator_instance = self.loop_generator()

        # Initialize a 'loop_id', and number of digits to use in it
        self.loop_id = "" if loop_id is None else loop_id
        # 'reverse' loop_id, showing remaining calls
        self.loop_id_r = "" if loop_id_r is None else loop_id_r
        self.digits = int(2 + np.log10(self.n_loops))

        # We save the raw input
        self.call_total = call_total
        self.call_number = call_number

    def loop_generator(self):
        for call_number, value in zip_longest(range(self.n_loops), self.values):
            # Custom print, as the debug_decorator prints don't work for the
            # generator setup. Also, here we loop over 'self.n_loops'
            self.info(f"Run {self.name} {call_number+1}/{self.n_loops}")

            # Update loop-id, to be used in filenames
            loop_id = self.loop_id + f"_{call_number:0{self.digits}d}"

            loop_id_r = self.loop_id_r + f"_{self.n_loops - call_number - 1:0{self.digits}d}"

            # Construct return_dict, and fix the call_total and call_number
            # in case of several loops.
            # call_total: multiply with length of the number of loops in this
            # call_number: add number of previous
            return_dict = {
                'call_total': self.call_total * self.n_loops,
                'call_number': call_number + self.call_number*self.n_loops,
                'loop_id': loop_id,
                'loop_id_r': loop_id_r,
            }

            # Possibly take new extent etc. from list
            try:
                extent = self.list_of_new_extents[call_number]
                return_dict['extent'] = extent
                return_dict['size'] = [extent[1] - extent[0], extent[3] - extent[2]]
            except AttributeError:
                pass

            if self.parameter is not None and value is not None:
                return_dict[self.parameter] = value

            yield return_dict


class EndLoop(Module):
    ''' Test... '''
    create_folder = False

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # Create a dict which will be used to merge the pipes from the
        # 'collected loops'
        self.new_pipe = {}

    def start(self, default=None, call_total=None, loop_id=None, loop_id_r=None,
              call_number=None, parameter=None, expression=None, values=None,
              overwrite=False,  # just to keep it from entering 'pipe'
              merge_function=None,  # TODO: Implement?
              **kwargs):
        # We save the raw input
        self.call_total = call_total

        self.loop_id = loop_id
        self.loop_id_r = loop_id_r

        self.loop_generator_instance = self.loop_generator()

        # We 'restart' with the call-number
        self.call_number = 0

        # Merge data
        for k, v in kwargs.items():
            self.merge_data(k, v)

    def merge_data(self, key, value):
        """
        Merges the data into self.new_pipe[key]. If the key already exists,
        it appends or concatenates the values accordingly.
        """
        data_to_merge = [
            'terrain_temp', 'terrain_prim',
            'position', 'width', 'height',
            'aspect', 'yaw_deg', 'pitch_deg',
            # Test:
            'roughness', 'heuristic_combined_roughness',
        ]

        if key in self.new_pipe and key in data_to_merge:
            if isinstance(self.new_pipe[key], np.ndarray) and isinstance(value, np.ndarray):
                # Use np.append() for numpy arrays
                self.new_pipe[key] = np.append(self.new_pipe[key], value, axis=0)
            elif isinstance(self.new_pipe[key], list) and isinstance(value, list):
                # Use list's extend() for lists
                self.new_pipe[key].extend(value)
            else:
                # For other types, just replace
                self.new_pipe[key] = value
        else:
            # If the key doesn't exist, simply assign the value
            self.new_pipe[key] = value

    def loop_generator(self):
        remaining_on_last_loop = int(self.loop_id_r.split("_")[-1])

        if remaining_on_last_loop != 0:
            # Return, i.e. don't be iterated over.
            # The effect is that the 'next module' will not be called,
            # and instead we return to the previous 'loop' module.
            # The idea is that we do this to 'collect' data, until the
            # last value from the loop. Then, we somehow 'merge'
            # all the collected data, and (finally) pass control along to the next module
            return
        else:
            # Calculate 'remaining' loop id
            loop_id = self.loop_id.rsplit('_', maxsplit=1)[0]
            loop_id_r = self.loop_id_r.rsplit('_', maxsplit=1)[0]

            # Construct a list of remaining iterations in each loop
            remaining = [int(id_r) for id_r in loop_id_r.split("_") if id_r]
            # Use this to calculated the remaining number of calls: call_total
            if len(remaining) > 0:
                call_total = np.prod(np.array(remaining)) + 1
            else:
                call_total = 1

            result = {
                'loop_id': loop_id,
                'loop_id_r': loop_id_r,
                'call_number': self.call_number,
                'call_total': call_total,
            }

            self.call_number += 1

            # Update result, and reset 'new_pipe'
            result = {**self.new_pipe, **result}
            # Reset
            self.new_pipe = {}

            yield result



from dataclasses import dataclass, field
import inspect
import os
import numpy as np


# ## Examples which are simply lists ###

# Flat ground
FLAT = [
    ('Plane', {}),
]


# Flat ground
ANGLED_PLANE = [
    ('Plane', {'pitch_deg': 10}),
]


# Gaussian hill, fully parameterized
GAUSSIAN_HILL = [
    ('Gaussian', {
        'position': [5, 5],
        'height': 2.5,
        'width': 5,
        'aspect': 1.5,
        'yaw_deg': 45,
        'pitch_deg': None,
    }),
]


# One way to get two Gaussian hills
GAUSSIAN_HILL_2 = [
    ('Gaussian', {
        'position': [[5, 5], [-5, 5]],
        'height': [2.5, 1.5],
        'width': 5,
        'aspect': 1.5,
        'yaw_deg': 45,
        'pitch_deg': None,
    }),
    ('Combine', {})
]

# Three random Gaussian hills
GAUSSIAN_HILL_RANDOM_3 = [
    ('Random', 3),
    ('Gaussian', {}),
    ('Combine', {})
]


# Large donut
DONUT_HILL = [
    ('Donut', {'width': 20}),
]


#
BASIC_AND_MAX = [
    ('Basic', [10, 10, 10, 10, 10]),
    ('Combine', 'Max'),
    ('Scale', 3),
]


BASIC_AND_MIN = [
    ('Basic', [10, 10, 10, 10, 10]),
    ('Combine', 'Min'),
    ('Scale', 3),
]


BASIC_AND_MAX2 = [
    ('Basic', [50, 20, 5]),
    ('Combine', 'Max'),
    ('Scale', 3),
]


BASIC_AND_MIN2 = [
    ('Basic', [50, 20, 5]),
    ('Combine', 'Min'),
    ('Scale', 3),
]


RANDOM_AND_SMOOTHSTEP = [
    ('SetDistribution', "height=uniform[0,2]"),
    ('Random', 5),
    ('SmoothStep', {}),
    ('Combine', 'Add'),
]


# Sample gaussian hills, along a circle
GAUSSIAN_HILLS_WITH_SAMPLED_LOCATIONS = [
    ('Donut', {'width': 20}),
    ('AsProbability', {}),  # Use donut as distribution,
    ('Random', 5),  # when sampling positions
    ('Gaussian', {'width': [2, 2, 2, 2, 2],  # We override the widths and heights
                  'height': [1, 1, 1, 1, 1]}),
    ('Combine', 'Max')  # combine with 'max', such that hills don't add to each other
]


# Mountain in horizon
MOUNTANS_IN_HORIZON = [
    ('Basic', 10),
    ('Add', 1.5),  # Get noise in [1.5, 2.5]
    ('Scale', 0.5),
    ('Donut', {'width': 30}),
    ('Combine', 'Prod')
]


# Explicit function of x
SINUS_X = [
    ('Function', 'np.sin(2*np.pi*x/20.0)'),
]


# Explicit function of r=np.sqrt(x^2+y^2)
SINUS_R = [
    ('Function', 'np.sin(2*np.pi*r/10.0)'),
]

# Explicit function of x and y
FUNCTION_XY = [
    ('Function', '5*(x/np.max(x))**2+np.sin(y/2)'),
]

# Rocks of different size
ROCKS = [
    ('Rocks', {}),
    ('Combine', {}),
]


# Rocks, unevenly scattered
SCATTERED_ROCKS = [
    ('Basic', 10),
    ('Scale', 10),
    ('Clip', {}),
    ('AsFactor', {}),
    ('Rocks', {}),
    ('Combine', {}),
    ('Scale', {}),
]


# Toy lunar surface example
LUNAR = [
    ('SetDistribution', 'height=uniform(0,0.5)'),
    ('SetDistribution', 'width=uniform(5,10)'),
    ('Random', 10),
    ('Donut', {}),
    ('Combine', 'Max'),
    ('Sphere', {'width': 500}),
    ('ToPrimary', {}),
    ('Combine', {}),
]


# Save to folder
SAVE = [
    ('Basic', 15),
    ('Save', 'Terrains/basic'),
]

# Load from folder
LOAD = [
    ('Load', 'Terrains/basic'),
]


# Set position/extent
EXTENT = [
    ('Extent', [-10, 20, -10, 20]),
    ('Plane', {}),
]


# Use Octaves
GROUND_FROM_NOISE = [
    ('Octaves', {}),
    ('WeightedSum', {}),
]


# Use Octaves, and save
GROUND_FROM_NOISE_SAVE = [
    ('Octaves', {}),
    ('Save', 'Terrains/octaves'),
    ('WeightedSum', {}),
]

# Load from single octaves, and construct different results
GROUND_FROM_NOISE_LOAD = [
    ('Load', 'Terrains/octaves'),
    ('Random', 'weights'),  # Special command to generate new weights
    ('WeightedSum', {}),
]


# Load from single octaves, and construct different results with set (approximate) slope and roughness
GROUND_FROM_NOISE_LOAD_SET_SLOPE_AND_ROUGHNESS = [
    ('Load', 'Terrains/octaves'),
    ('Random', 'weights'),  # Special command to generate new weights
    ('Slope', {}),
    ('SetSlope', 5),
    ('Roughness', {}),
    ('SetRoughness', 1.03),
    ('WeightedSum', {}),
]


# Sample slope from some distribution
GROUND_FROM_NOISE_LOAD_SAMPLE_SLOPE = [
    ('Load', 'Terrains/octaves'),
    ('Random', 'weights'),  # Special command to generate new weights
    ('Slope', {}),
    ('SetDistribution', 'target_slope_deg=uniform(0,10)'),
    ('Sample', 'target_slope_deg'),
    ('SetSlope', {}),
    ('Roughness', {}),
    ('SetRoughness', 1.03),
    ('WeightedSum', {}),
]

# Use Loop:2x2 and Stack.
# Could be used for parallelization, but is currently not
SPLIT_AND_STACK = [
    ('GridSize', 150),  # Hack, must half this
    ('Loop', '2x2'),
    ('Seed', 'persistent_random'),  # Get same seed each loop, but different each run
    ('Basic', {}),
    ('WeightedSum', {}),
    ('EndLoop', {}),
    ('Stack', {}),
]


# Use Over to make flat at machine
FLAT_IN_CENTER = [
    ('Basic', {}),
    ('WeightedSum', {}),
    ('Cube', {'width': 5}),
    ('AsFactor', {}),
    ('Plane', {}),
    ('ToPrimary', {}),
    ('Compose', {}),
]


# Classes to make 'configs' (Cfg) with default, changeable parameters
@dataclass
class ArtificialTerrainCfg:
    @property
    def modules(self) -> list:
        """Each config must expose a list of modules."""
        raise NotImplementedError(f"{type(self).__name__} must define a `modules` property")
        # return []

    # Needs to be iterable to behave as a list for artifical_terrains
    def __iter__(self):
        return iter(self.modules)

    # Make sure cfg:s can be added, with the result being a 'Combined cfg'
    def __add__(self, other: "ArtificialTerrainCfg") -> "CombinedArtificialTerrainCfg":
        if not isinstance(other, (ArtificialTerrainCfg, list)):
            return NotImplemented
        return CombinedArtificialTerrainCfg(children=[self, other])

    def __radd__(self, other):
        if not isinstance(other, (ArtificialTerrainCfg, list)):
            return NotImplemented
        return CombinedArtificialTerrainCfg(children=[other, self])


@dataclass
class CombinedArtificialTerrainCfg(ArtificialTerrainCfg):
    """Composite config, holds children configs."""
    children: list = field(default_factory=list)

    @property
    def modules(self):
        # flatten children’s modules
        combined = []
        for child in self.children:
            if isinstance(child, list):
                combined.extend(child)
            else:
                combined.extend(child.modules)
        return combined

    def __setattr__(self, key, value):
        """Try to propagate sets into children if they have the attribute."""
        if key in {"children", "modules"}:
            # Don't redirect for these two special attributes.
            # They are "structural" parts of the composite.
            object.__setattr__(self, key, value)
            return
        else:
            # propagate to children only
            for child in self.children:
                if hasattr(child, key):
                    setattr(child, key, value)


@dataclass
class FlatCfg(ArtificialTerrainCfg):
    @property
    def modules(self):
        return [
            ('Plane', {}),
        ]


@dataclass
class AngledPlaneCfg(ArtificialTerrainCfg):
    pitch_deg: float = 10

    @property
    def modules(self):
        return [
            ('Plane', {'pitch_deg': self.pitch_deg}),
        ]


# Gaussian hill, fully specified
@dataclass
class GaussianHillCfg(ArtificialTerrainCfg):
    @property
    def modules(self):
        return [
            ('Gaussian', {
                'position': [5, 5],
                'height': 2.5,
                'width': 5,
                'aspect': 1.5,
                'yaw_deg': 45,
                'pitch_deg': None,
            }),
        ]


# N number of random Gaussian hills
@dataclass
class GaussianHillRandomCfg(ArtificialTerrainCfg):
    number: int = 3

    @property
    def modules(self):
        return [
            ('Random', self.number),
            ('Gaussian', {}),
            ('Combine', {}),
            # Allowing adding to other:
            # (even if that terrain is in 'temporary')
            ('ToPrimary', {}),
            ('Combine', {}),
        ]


# Mountain in horizon
@dataclass
class MountansInHorizonCfg(ArtificialTerrainCfg):
    width: float = 30

    @property
    def modules(self):
        return [
            ('Basic', 10),
            ('Add', 1.5),  # Get noise in [1.5, 2.5]
            ('Scale', 0.5),
            ('Donut', {'width': self.width}),
            ('Combine', 'Prod'),
            # Allowing adding to other:
            ('ToPrimary', {}),
            ('Combine', {}),
        ]


# Explicit function
@dataclass
class ExplicitFunctionCfg(ArtificialTerrainCfg):
    function: str = 'np.sin(2*np.pi*x/20.0)'

    @property
    def modules(self):
        return [
            ('Function', self.function),
        ]


@dataclass
class LoadOrGenerateAndSetSlopeCfg(ArtificialTerrainCfg):
    ''' Example of some extended functionality'''
    folder: str = 'Terrains/octaves'
    slope_deg: float = 8
    roughness: float = 1.03

    @property
    def modules(self):
        modules = []
        if os.path.exists(self.folder):
            modules.extend(
                [('Load', self.folder),
                 ('Random', 'weights'),]
            )
        else:
            modules.extend(
                [('Octaves', {}),
                 ('Save', self.folder),]
            )
        modules.extend(
            [
                ('Slope', {}),
                ('SetSlope', self.slope_deg),
                ('Roughness', {}),
                ('SetRoughness', self.roughness),
                ('WeightedSum', {}),
            ]
        )
        return modules


@dataclass
class LoadFromSeveralAndSetSlopeAndRoughnessCfg(ArtificialTerrainCfg):
    ''' Example of some extended functionality'''
    folder: str = 'Terrains/multiple_octaves'
    number_of_sets: int = 5
    slope_deg: float = 8
    roughness: float = 1.05

    @property
    def modules(self):
        import random
        digits = int(2 + np.log10(self.number_of_sets))

        modules = []
        # Define filename to check that files exist
        a_filename = f'terrain_temp_{self.number_of_sets - 1:0{digits}d}_{0:05d}.npz'
        if not os.path.exists(os.path.join(self.folder, a_filename)):
            modules.extend(
                [('Loop', self.number_of_sets),
                 ('Octaves', {}),
                 ('Save', self.folder),
                 ('ClearTerrain', None),
                 ('EndLoop', None),
                 ]
            )

        files = [
            f'{self.folder}/terrain_temp_{random.randint(0, self.number_of_sets - 1):0{digits}d}_{i:05d}.npz'
            for i in range(10)
        ]
        modules.extend(
            [
                ('Load', files),
                ('Random', 'weights'),
                ('Slope', {}),
                ('SetSlope', self.slope_deg),
                ('Roughness', {}),
                ('SetRoughness', self.roughness),
                ('WeightedSum', {}),
            ]
        )
        return modules


@dataclass
class LoadFromSeveralAndSetSlopeCfg(ArtificialTerrainCfg):
    ''' Example of some extended functionality'''

    num_octaves: int = 10
    start: float = 128

    folder: str = 'Terrains/multiple_octaves2'
    number_of_sets: int = 5
    slope_deg: float = 8

    @property
    def modules(self):
        import random
        digits = int(2 + np.log10(self.number_of_sets))

        modules = []
        # Define filename to check that files exist
        a_filename = f'terrain_temp_{self.number_of_sets - 1:0{digits}d}_{0:05d}.npz'
        if not os.path.exists(os.path.join(self.folder, a_filename)):
            modules.extend(
                [('Loop', self.number_of_sets),
                 ('Octaves', {'num_octaves': self.num_octaves, 'start': self.start}),
                 ('Save', self.folder),
                 ('ClearTerrain', None),
                 ('EndLoop', None),
                 ]
            )

        files = [
            f'{self.folder}/terrain_temp_{random.randint(0, self.number_of_sets - 1):0{digits}d}_{i:05d}.npz'
            for i in range(self.num_octaves)
        ]
        modules.extend(
            [
                ('Load', files),
                ('Random', 'weights'),
                ('Slope', {}),
                ('SetSlope', self.slope_deg),
                ('WeightedSum', {}),
            ]
        )
        return modules


@dataclass
class AndFlatInCenterCfg(ArtificialTerrainCfg):
    width: float = 5

    @property
    def modules(self):
        return [
            # Assumes there is already a terrain in primary
            ('Cube', {'width': self.width}),
            ('AsFactor', {}),
            ('Plane', {}),
            ('ToPrimary', {}),
            ('Compose', {}),
        ]


@dataclass
class AndSetDifficulty(ArtificialTerrainCfg):
    ''' Curriculum thing, interpolating with Plane to set difficulty '''
    difficulty: float = 0

    @property
    def modules(self):
        return [
            # Assumes there is already a terrain in primary
            ('Set', f'factor={self.difficulty}'),
            ('Plane', {}),
            ('ToPrimary', {}),
            ('Compose', 'Under')
        ]


@dataclass
class StoneCircleCfg(ArtificialTerrainCfg):
    width: float = 25
    number_of_stones: int = 30

    @property
    def modules(self):
        return [
            ('Donut', {'width': self.width}),
            ('AsProbability', {}),
            ('SetDistribution', "width=uniform(2,2)"),
            ('Random', self.number_of_stones),
            ('Cube', {}),
            ('Combine', "Max"),
            # Allowing adding to other:
            ('ToPrimary', {}),
            ('Combine', {}),
        ]


@dataclass
class PerimiterWallsCfg(ArtificialTerrainCfg):
    width: float = 28.0
    height: float = 2.0

    @property
    def modules(self):
        return [
            ('Cube', {'width': self.width, 'height': self.height}),
            ('Negate', {}),
            ('Add', self.height),
            # Allowing adding to other:
            ('ToPrimary', {}),
            ('Combine', {}),
        ]


@dataclass
class TerrainAndHillsAndPerimiterCfg(ArtificialTerrainCfg):
    ''' Combine several things into something semi-complicated'''

    # Size and grid-size
    size_x: float = 50
    size_y: float = 50
    grid_size_x: int = 500
    grid_size_y: int = 500
    # Parameters for general terrain
    slope_deg: float = 5

    folder: str = 'Terrains/combine_several'

    perimiter_walls: bool = True
    gaussian_hills: int = 10

    def __post_init__(self):
        # Add hash to self.folder, to get unique folders for different
        # size and grid-size

        import hashlib
        # Compute hash from instance values
        data = (self.size_x, self.size_y, self.grid_size_x, self.grid_size_y)
        data_bytes = repr(data).encode("utf-8")
        hash_digest = hashlib.sha1(data_bytes).hexdigest()[:6]

        self.folder = f"{self.folder}_{hash_digest}"

    @property
    def modules(self):

        # Add hash to folder
        print(f"self.folder:{self.folder}")
        # exit(0)

        # Calculate number of octaves and start
        # We want 'start' a factor 2~4 more than the max size,
        # and then set num_octaves to get the smallest features
        # in the order of 1 meter.
        max_size = np.maximum(self.size_x, self.size_y)
        start_log = int(np.log2(max_size)+2)
        end_log = 1
        num_octaves = start_log - end_log + 1
        start = np.power(2, start_log)

        size = [
            # Set Size and GridSize
            ('Size', [self.size_x, self.size_y]),
            ('GridSize', [self.grid_size_x, self.grid_size_y]),
        ]

        # Setup general terrain
        terrain = LoadFromSeveralAndSetSlopeCfg(
            folder=self.folder,
            slope_deg=self.slope_deg,
            num_octaves=num_octaves,
            start=start,
        )

        # Setup walls
        if self.perimiter_walls:
            perimiter = PerimiterWallsCfg(width=self.size_x-2)
        else:
            perimiter = []

        if self.gaussian_hills:
            hills = GaussianHillRandomCfg(number=self.gaussian_hills)
        else:
            hills = []

        return size + terrain + perimiter + hills


COMBINED_SET_SLOPE_AND_FLAT_IN_CENTER = (
    LoadOrGenerateAndSetSlopeCfg()
    + AndFlatInCenterCfg()
)


COMBINED_SET_SLOPE_AND_PERIMITER_WALLS = (
    LoadOrGenerateAndSetSlopeCfg()
    + PerimiterWallsCfg()
)


COMBINED_SET_SLOPE_AND_STONE_CIRCLE = (
    LoadOrGenerateAndSetSlopeCfg()
    + StoneCircleCfg()
)


COMBINED_SET_SLOPE_AND_MOUNTAINS_IN_HORIZON = (
    LoadOrGenerateAndSetSlopeCfg()
    + MountansInHorizonCfg()
)


COMBINED_SET_SLOPE_AND_GAUSSIAN_HILLS = (
    LoadOrGenerateAndSetSlopeCfg()
    + GaussianHillRandomCfg()
)


INSTANTIATED_CFG_GAUSSIAN_HILLS = GaussianHillRandomCfg(number=10)

INSTANTIATED_CFG_TERRAIN_HILL_PERIMITER_SLOPE_5 = TerrainAndHillsAndPerimiterCfg()
INSTANTIATED_CFG_TERRAIN_HILL_PERIMITER_SLOPE_10 = TerrainAndHillsAndPerimiterCfg(
    slope_deg=10)
INSTANTIATED_CFG_TERRAIN_HILL = TerrainAndHillsAndPerimiterCfg(
    perimiter_walls=False
)
INSTANTIATED_CFG_TERRAIN_PERIMITER = TerrainAndHillsAndPerimiterCfg(
    gaussian_hills=0
)
INSTANTIATED_CFG_TERRAIN = TerrainAndHillsAndPerimiterCfg(
    gaussian_hills=0, perimiter_walls=False
)

INSTANTIATED_CFG_TERRAIN_STRIP_Y = TerrainAndHillsAndPerimiterCfg(
    gaussian_hills=0, perimiter_walls=False, size_x=10, grid_size_x=100,
)
INSTANTIATED_CFG_TERRAIN_STRIP_X = TerrainAndHillsAndPerimiterCfg(
    gaussian_hills=0, perimiter_walls=False, size_y=10, grid_size_y=100,
)


# 10 different terrains from the same set of octaves


if __name__ == "__main__":
    """
    Script entry point for plot/renders of example configurations

    Usage:
    Run the script directly with Python:
            python -m artificial_terrains.examples

    # Note: for the reason of my probably not handling all imports
    # correctly, the following does not work:
    python artificial_terrains_cfg.py

    Or run it inside Blender in background mode:
        blender --python artificial_terrains_cfg.py --background

    Notes:
    - Rendering requires Blender and artificial_terrains package
    """

    current_module = globals()
    configs = {
        name: value for name, value in current_module.items()
        if name.isupper() and isinstance(value, list)
        and not name.startswith("AND")
    }

    # other style: dataclass config with a .modules property
    cfg_configs = {
        name: value() for name, value in current_module.items()
        if inspect.isclass(value) and issubclass(value, ArtificialTerrainCfg)
        and not name.startswith("And")
    }
    del cfg_configs['ArtificialTerrainCfg']
    del cfg_configs['CombinedArtificialTerrainCfg']

    combined_configs = {
        name: value for name, value in current_module.items()
        if name.isupper() and isinstance(value, CombinedArtificialTerrainCfg)
    }

    instantiated_configs = {
        name: value for name, value in current_module.items()
        if name.isupper() and isinstance(value, ArtificialTerrainCfg)
    }

    # for name, _class in cfg_configs.items():
    configs = {
        **configs,
        **cfg_configs,
        **combined_configs,
        **instantiated_configs,
    }

    # Override with test
    # configs = {'test': LoadFromSeveralAndSetSlopeCfg()}

    for name, config in configs.items():
        print(f"\n=== Running config: {name} ===")
        import artificial_terrains as at

        base_cfg = [
            ('Size', [30, 30]),
            ('GridSize', [300, 300]),
        ]

        plot_cfg = [
            ('Plot', {'filename_base': f'{name}',
                      'folder': 'example_plots'}),
        ]

        folder = 'example_renders'
        filename = f'{name}.png'

        render_cfg = [
            ('ClearScene', None),
            ('BasicSetup', None),
            ('Holdout', None),
            ('Ground', {}),
            ('Camera', 45),
            ('Render', {'folder': folder,
                        'filename': filename}),
        ]

        # Possibly skip already done
        if os.path.isfile(os.path.join(folder, filename)):
            continue

        # Try to run as blender script
        try:
            at.run(base_cfg + config + render_cfg + plot_cfg, verbose=True)
        except ModuleNotFoundError:
            # Otherwise just run as a plot script
            at.run(base_cfg + config + plot_cfg, verbose=True)



import argparse
import sys
import yaml
import os
import re


def create_parser(parser=None, specific_args=[], positional=[]):
    if parser is None:
        parser = argparse.ArgumentParser()

    # Common arguments
    parser.add_argument(
        '--save-dir', type=str, default=None, required=True,
        help='Explicit save dir')
    parser.add_argument(
        '--settings-file', type=str, default=None,
        help='File to load Env settings from')
    parser.add_argument(
        '--settings', type=str, nargs='+', action=StoreDict,
        help='Overwrite settings (e.g. log:cylinder pile:single)'
    )
    parser.add_argument(
        '--debug', type=int, choices=[0, 1, 2, 3], default=2,
        help='Set debug level: 0 (NOTSET), 1 (DEBUG), 2 (INFO), 3 (WARNING)')

    # Handle specific arguments
    for name in specific_args:
        if name == 'modules':
            parser.add_argument(
                '--modules', type=str, help='Modules to use',
                action=StoreTuplePair,
                default=(None,), nargs='+', required=True,
            )

    return parser


def parse_options(options):
    ''' parse 'options' to make it a dict '''
    # Extract possible settings kwargs
    if isinstance(options, dict):
        kwargs = options
    elif options is None:
        kwargs = {}
    else:
        # Store value as 'default'
        kwargs = {'default': options}

    return kwargs


def get_args_combined_with_settings(parser):
    '''
    Combine args with settings, complementing argparse from settings-file

    Input can be given via
     1) a yaml settings file,
     2) a settings dict, or
     3) argparse argument

    # Examples

    1): python script.py --settings-file with
    spawn-type:
      AtRock:
        num_spawns: 1
        plot: True

    (not designed for TuplePairs, misses information here)
    2): python script --settings spawn-type:AtRock

    3) python script.py --spawn-type AtRockEdge:"dict(num_spawns=1)"

    Priorities, from lowest to highest
      --settings-file
      --settings
      --argparse argument
    '''
    # Get args
    args, _ = parser.parse_known_args()

    # Make args dict
    args_dict = vars(args)

    # Construct a mapping between args-names and 'actions'
    # To know if an argument is a StoreDict, TuplePair
    argument_action_mapping = {}
    for action in parser._actions:
        try:
            action_name = type(action).__name__
            name = action.option_strings[0].replace('--', '').replace('-', '_')
            argument_action_mapping[name] = action_name
        except IndexError:
            pass

    # Construct combined 'settings' from settings-file and --settings
    settings = load_settings(args.settings_file, args)

    # Possibly take options from settings?
    for args_name, args_value in args_dict.items():
        lowercase_name = args_name
        dash_name = f"{lowercase_name.replace('_', '-')}"
        from_commandline = f'--{dash_name}' in sys.argv[1:]
        # Get action type
        action = None
        if lowercase_name in argument_action_mapping:
            action = argument_action_mapping[lowercase_name]

        # If in settings, and not explicitly given via command line
        # --> take from settings
        if not from_commandline and dash_name in settings:
            settings_value = settings[dash_name]

            # Handle TuplePairs
            if action == 'StoreTuplePair':
                if type(settings_value) is str:
                    tuple_pair = [(settings_value, None)]
                elif type(settings_value) is dict:
                    tuple_pair = []
                    for k, v in settings_value.items():
                        tuple_pair.append((k, v))
                        # tuple_pair.append([k, v])
                elif type(settings_value) is list:
                    tuple_pair = settings_value

                setattr(args, lowercase_name, tuple_pair)
            if action == '_StoreAction':
                setattr(args, lowercase_name, settings_value)
            elif action == 'StoreDict':
                print(f"action:{action}")
                setattr(args, lowercase_name, settings_value)
            elif action is None:
                setattr(args, lowercase_name, settings_value)

        # if explicitly given via command line
        # --> add to settings
        if from_commandline:
            if action == '_StoreAction':
                settings[dash_name] = args_value
            elif action == 'StoreTuplePair':
                # Change from list of tuple, to dict.
                # From: [('AtRock', {'num_spawns': 100})]
                # To:   {'AtRock': {'num_spawns': 100}}

                # Fix tuple to list
                args_value = [list(t) for t in args_value]
                # settings[dash_name] = {args_value[0][0]: args_value[0][1]}
                # Append to list if already in settings and is a list.
                if dash_name in settings and isinstance(settings[dash_name], list):
                    settings[dash_name].extend(args_value)
                else:
                    # other wise we just set the value
                    settings[dash_name] = args_value
            elif action == 'StoreDict':
                # TODO: Not tested
                settings[dash_name] = args_value
            elif action is None:
                # TODO: Not tested
                settings[dash_name] = args_value

        # if not in settings
        # --> add to settings
        if dash_name not in settings:
            settings[dash_name] = args_value

    # Possibly remove 'settings-file' from settings to avoid recursion
    if 'settings-file' in settings:
        del settings['settings-file']
    if 'settings' in settings:
        del settings['settings']

    # Create save_dir
    if args.save_dir and not os.path.exists(args.save_dir):
        os.makedirs(args.save_dir)

    # ## module specific code ###
    # if not [name, option] pair, (only name)
    from modules.modules import MODULES
    for i, pair in enumerate(args.modules):
        if not isinstance(pair, (tuple, list)):
            key = pair
            if key in MODULES.keys():
                args.modules[i] = [key, None]

    # Save settings
    if args.save_dir is not None:
        filename = os.path.join(args.save_dir, 'settings.yml')

        # ## module specific code ###
        # Process modules to simplify yaml file
        # Remove [name, None] and turn to 'name'
        if 'modules' in settings:
            # Process settings to remove unnecessary null's
            processed_modules = [
                [name, options] if options is not None else
                name for name, options in settings['modules']]
            settings['modules'] = processed_modules

        with open(filename, 'w') as f:
            yaml.dump(settings, f, sort_keys=False, default_flow_style=None)

    return args


def load_settings(yaml_file, args):
    import yaml

    # Yaml bug workaround,
    # https://stackoverflow.com/questions/30458977/yaml-loads-5e-6-as-string-and-not-a-number
    # (seem to be work to add this to yaml main)
    loader = yaml.SafeLoader
    loader.add_implicit_resolver(
        u'tag:yaml.org,2002:float',
        re.compile(u'''^(?:
     [-+]?(?:[0-9][0-9_]*)\\.[0-9_]*(?:[eE][-+]?[0-9]+)?
    |[-+]?(?:[0-9][0-9_]*)(?:[eE][-+]?[0-9]+)
    |\\.[0-9_]+(?:[eE][-+][0-9]+)?
    |[-+]?[0-9][0-9_]*(?::[0-5]?[0-9])+\\.[0-9_]*
    |[-+]?\\.(?:inf|Inf|INF)
    |\\.(?:nan|NaN|NAN))$''', re.X),
        list(u'-+0123456789.'))

    if yaml_file is not None:
        # Load settings from yaml file
        with open(yaml_file, 'r') as f:
            # settings_dict = yaml.safe_load(f)
            settings = yaml.load(f, Loader=loader)
            # if env_str in list(settings_dict.keys()):
            #     settings = settings_dict[env_str]
            # else:
            #     # raise ValueError("Settings not found for {}".format(env_str))
            #     settings = settings_dict
    else:
        # Else populate with empty dict
        settings = {}

    # Overwrite hyperparams from command line, if exists
    if args is not None and args.settings is not None:
        for k, v in args.settings.items():
            if type(v) is not dict:
                settings[k] = v
            # Solution for overwriting single kwargs of second 'level'
            elif k not in settings:
                settings[k] = v
            else:
                for k2, v2 in v.items():
                    settings[k][k2] = v2

    return settings


class StoreDict(argparse.Action):
    """
    Custom argparse action for storing dict.

    In: args1:0.0 args2:"dict(a=1)"
    Out: {'args1': 0.0, arg2: dict(a=1)}
    """
    def __init__(self, option_strings, dest, nargs=None, **kwargs):
        self._nargs = nargs
        super(StoreDict, self).__init__(option_strings,
                                        dest,
                                        nargs=nargs,
                                        **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        arg_dict = {}
        for arguments in values:
            arg_dict_local = self.split(arguments)
            arg_dict = {**arg_dict, **arg_dict_local}
        setattr(namespace, self.dest, arg_dict)

    def split(self, arguments):
        arg_dict = {}
        key = arguments.split(":")[0]
        value = ":".join(arguments.split(":")[1:])
        # Evaluate the string as python code
        try:
            if ':' in value:
                arg_dict_lower = self.split(value)
                arg_dict[key] = arg_dict_lower
            else:
                arg_dict[key] = eval(value)
        except NameError:
            arg_dict[key] = value
        except SyntaxError:
            return {key: value}

        return arg_dict


class StoreTuplePair(argparse.Action):
    """
    Custom argparse action for storing tuple pair.

    In:
    Out:
    """
    def __init__(self, option_strings, dest, nargs=None, **kwargs):
        self._nargs = nargs
        super(StoreTuplePair, self).__init__(
            option_strings,
            dest,
            nargs=nargs,
            **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        # Store list of tuple-pairs
        tuple_pair_list = []
        for arguments in values:
            key = arguments.split(":")[0]
            value = ":".join(arguments.split(":")[1:])
            # Evaluate the string as python code
            try:
                tuple_pair_list.append((key, eval(value)))
            except NameError:
                tuple_pair_list.append((key, value))
            except SyntaxError:
                if value == "":
                    tuple_pair_list.append((key, None))
                else:
                    tuple_pair_list.append((key, value))
        setattr(namespace, self.dest, tuple_pair_list)


class StoreTuplePair2(argparse.Action):
    """
    Another way of implementing the TuplePair.
    I'm not sure if they are different...

    In: args1:0.0 args2:"dict(a=1)"
    Out: {'args1': 0.0, arg2: dict(a=1)}
    """
    def __init__(self, option_strings, dest, nargs=None, **kwargs):
        self._nargs = nargs
        super(StoreTuplePair2, self).__init__(
            option_strings,
            dest,
            nargs=nargs,
            **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        # arg_dict = {}
        tuple_pair_list = []
        for arguments in values:
            arg_dict_local = self.split(arguments)
            for key, value in arg_dict_local.items():
                tuple_pair_list.append((key, value))

        setattr(namespace, self.dest, tuple_pair_list)

    def split(self, arguments):
        arg_dict = {}
        key = arguments.split(":")[0]
        value = ":".join(arguments.split(":")[1:])
        # Evaluate the string as python code
        try:
            if ':' in value:
                arg_dict_lower = self.split(value)
                arg_dict[key] = arg_dict_lower
            else:
                arg_dict[key] = eval(value)
        except NameError:
            arg_dict[key] = value
        except SyntaxError:
            return {key: value}

        return arg_dict



import numpy as np
import matplotlib.pyplot as plt


psi_deg = np.arange(0, 360)
psi_rad = psi_deg*np.pi/180

psi_0 = np.zeros_like(psi_deg)

theta_deg = 20
theta_rad = theta_deg*np.pi/180

# We consider two planes with pitch 20 degrees.
# We keep one fix, and rotate the other 360 degrees
# Add them for each configuration, and calculate the resultant angle

new_angle = np.tan(theta_rad)*np.abs(2*np.cos(psi_rad - psi_0))
new_angle_deg = new_angle*180/np.pi

fig, ax = plt.subplots(figsize=(8, 4))
ax.plot(new_angle_deg)

# ax.set_title("Histogram of Mean Slope (degrees)")
ax.set_xlabel("x-value (a.u)")
ax.set_ylabel("Mean slope degree")
ax.grid()

fig.tight_layout()

# Save the figure
save_path = "test2.png"
fig.savefig(save_path)
plt.close(fig)




from utils.plots import debug_plot
import os


global plot_debug_counter
plot_debug_counter = 0
plot_debug_enabled = False


def debug(module, kwargs, call_number, call_total, name):
    global plot_debug_enabled, plot_debug_counter

    if not plot_debug_enabled or kwargs is None:
        return

    # Plot input
    filename = f'debug_{plot_debug_counter:05d}_{name}_{module.name}_{call_number+1}_{call_total}.png'
    filename = os.path.join(module.save_dir_original, filename)
    debug_plot(filename, **kwargs)
    plot_debug_counter += 1



# Expose a friendly API to use as a library

"""
Artificial Terrains
===================

Modular, extensible terrain generation library.

"""
import sys
import os
import logging

# make artificial_terrains.examples as a valid import:
from .utils import examples
__all__ = ["examples"]

# Hack: extend sys.path so imports still work after moving all code into
# artificial_terrains/artificial_terrains for packaging. This avoids having
# to rewrite all existing `import modules...` calls right now.
sys.path.append(os.path.dirname(__file__))


def run(modules, save_dir="", logger=None, general_kwargs=None, pipe=None,
        verbose=False):
    """
    Run a list of (name, options) modules just like the CLI script.
    Example:
        run([("Basic", {}), ("WeightedSum", {})])
    """

    global _last_pipe

    if logger is None:
        # Always use the same base logger name
        logger = logging.getLogger("artificial_terrains")
        # Essentially turn of logging by default
        logger.setLevel(logging.CRITICAL)

    if verbose:
        # Clear any logger to avoid doubles
        logger.handlers.clear()

        # Add a logger printing to stdout
        handler = logging.StreamHandler(sys.stdout)
        # formatter = logging.Formatter("%(levelname)s: %(message)s")
        # handler.setFormatter(formatter)
        logger.addHandler(handler)
        logger.setLevel(logging.INFO)

    if general_kwargs is None:
        general_kwargs = {}

    if pipe is None:
        pipe = {
            'size': 50,
            'resolution': 2,
            'extent': [-25, 25, -25, 25],
            'call_number': 0,
            'call_total': 1,
        }

    from modules.modules import MODULES
    # Build module objects
    list_of_modules_kwargs_tuples = []
    for module, options in modules:
        if module is None:
            continue
        module_class = MODULES[module]
        module_obj = module_class(save_dir, logger)
        if isinstance(options, dict):
            kwargs = options
        elif options is None:
            kwargs = {}
        else:
            kwargs = {'default': options}
        kwargs = {**general_kwargs, **kwargs}
        list_of_modules_kwargs_tuples.append((module_obj, kwargs))

    # Run recursively
    from .generate_terrain import recursive_module_call
    last_pipe = recursive_module_call(
        list_of_modules_kwargs_tuples, pipe=pipe, logger=logger)

    # Store last_pipe globally
    _last_pipe = last_pipe


def get_terrain():
    """Return the last terrain (single)."""

    if 'terrain_prim' in _last_pipe:
        return _last_pipe['terrain_prim'][-1]
    elif 'terrain_temp' in _last_pipe:
        return _last_pipe['terrain_temp'][-1]
    else:
        return None


def get_terrains():
    """Return all terrains from the last run."""

    if 'terrain_prim' in _last_pipe:
        return _last_pipe['terrain_prim']
    elif 'terrain_temp' in _last_pipe:
        return _last_pipe['terrain_temp']
    else:
        return None



import logging
import sys
from datetime import datetime
import os


# Map debug levels to logging levels
level_map = {
    0: logging.NOTSET,
    1: logging.DEBUG,
    2: logging.INFO,
    3: logging.WARNING
}


def get_logger(name='logger', save_dir=None, level=logging.INFO):
    '''
    Create logger with stdout- (and possibly file-) handler
    '''
    # Make sure directory exists
    if save_dir and not os.path.exists(save_dir):
        os.makedirs(save_dir)

    # Construct filename
    if save_dir is not None:
        filename = os.path.join(save_dir, f'{name}.txt')
    else:
        filename = None

    logger = logging.getLogger(name)
    # Add stdout handler
    stream = logging.StreamHandler(sys.stdout)
    stream.setLevel(logging.DEBUG)  # ?
    logger.addHandler(stream)
    if filename is not None:
        # Add file handler
        hdlr = logging.FileHandler(filename)
        logger.addHandler(hdlr)
    # Set overall level
    logger.setLevel(level)

    # Output command line input
    command_line = ' '.join(sys.argv)
    command_line = parse_command(command_line)

    # workaround
    if 'python blender.py' in command_line:
        command_line = command_line.replace(
            'python blender.py',
            'blender --python blender.py --')

    logger.info(f'### {command_line}')
    logger.info(datetime.now())

    return logger


def parse_command(command_line):
    ''' Fix things disappeared from command line string

    Add double qoutes around dict(..)
    Add 'python' to start
    '''
    import re
    pattern = r"dict\([^)]+\)"

    def add_quotation_marks(match):
        matched_string = match.group(0)
        return f'"{matched_string}"'

    # Replace occurrences of dict(...) with 'dict(...)'
    modified_string = re.sub(pattern, add_quotation_marks, command_line)
    modified_string = 'python ' + modified_string

    return modified_string



import numpy as np
from scipy.interpolate import RegularGridInterpolator


class Interpolator():
    """
    Sample from heightfield
    """
    def __init__(self, hf_array, extent):
        # Setup the interpolator using the hf_array and extent
        self.setup_interpolator(hf_array, extent)

    def setup_interpolator(
            self, hf_array,
            extent,
            **_,
    ):
        left, right, bottom, top = extent

        x = np.linspace(left, right, hf_array.shape[0])
        y = np.linspace(bottom, top, hf_array.shape[1])

        self.interpolator = RegularGridInterpolator(
            (x, y),
            hf_array,
            method='linear',
            bounds_error=False,
            fill_value=None)

    def __call__(self, x, y):
        """
        Computes the height at a given position.

        Args:
            x (float): The x-coordinate of the position.
            y (float): The y-coordinate of the position.

        Returns:
            float: The height at the given position.
        """
        result = self.interpolator([x, y])
        return result

    def setup_points(self, size=(2, 2), grid_size=(128, 128), offset_x=0):
        '''
        Setup points for sampling heightfield
        '''
        half_x = size[0] / 2
        half_y = size[1] / 2
        nx, ny = grid_size
        self.grid_size = grid_size
        x = offset_x + np.linspace(-half_x, half_x, nx)
        y = np.linspace(-half_y, half_y, ny)

        return np.array(np.meshgrid(x, y)).T.reshape(-1, 2)

    def get_height_map(self, pos: tuple, theta: float, points):
        '''
        Return a heightmap, from the 'points', positioned at 'pos'
        and angled in 'theta'.
        '''
        def rotmat(theta):
            return np.array([[np.cos(theta), -np.sin(theta)],
                             [np.sin(theta), np.cos(theta)]])

        # Transform grid to front body frame in world coordinates
        trans_grid_points = rotmat(theta).dot(points.T).T
        trans_grid_points += np.array(pos)

        heights = self.interpolator(trans_grid_points)

        # Reshape to grid size
        heights = heights.reshape(self.grid_size)
        self.heights = heights

        return heights, trans_grid_points



import bpy
import numpy as np
import os


def setup_near_far():
    for area in bpy.context.screen.areas:
        if area.type == 'VIEW_3D':
            for space in area.spaces:
                if space.type == 'VIEW_3D':
                    space.clip_start = 0.1
                    space.clip_end = 5000   # <-- set far clip here


def remove_object(name='Cube'):
    if name in bpy.data.objects:
        grid_to_remove = bpy.data.objects[name]
        bpy.data.objects.remove(grid_to_remove, do_unlink=True)

        return True
    return False


def get_material(name="GroundMaterial"):
    ''' Create or get a material '''
    if name not in bpy.data.materials:
        material = bpy.data.materials.new(name=name)
    else:
        material = bpy.data.materials.get(name)

    return material


def setup_z_coord_shader(material):
    ''' Setup shader with colormap for z value '''
    # Setup to use nodes
    material.use_nodes = True

    # Get material node tree and nodes
    node_tree = material.node_tree
    nodes = node_tree.nodes

    # Remove all previous nodes
    for node in nodes:
        nodes.remove(node)

    # Dict with shaders to create
    nodes_to_add = {
        'texture': 'ShaderNodeTexCoord',
        'separate': 'ShaderNodeSeparateXYZ',
        # 'multiply': 'ShaderNodeMath',
        # 'maprange': 'ShaderNodeMapRange',
        'colorramp': 'ShaderNodeValToRGB',
        'bsdf': 'ShaderNodeBsdfPrincipled',
        'output': 'ShaderNodeOutputMaterial',
    }

    # Create shaders
    added_nodes = {}
    x_location = -600
    for name, shader in nodes_to_add.items():
        node = nodes.new(shader)
        # Set position
        node.location.x = x_location
        node.location.y = 200
        x_location += 250
        # Add to dict
        added_nodes[name] = node

    # Make shorter names for convenience
    texture_node = added_nodes['texture']
    separate_node = added_nodes['separate']
    # multiply_node = added_nodes['multiply']
    # maprange_node = added_nodes['maprange']
    colorramp_node = added_nodes['colorramp']
    bsdf_node = added_nodes['bsdf']
    output_node = added_nodes['output']

    # Setup some nodes
    # multiply_node.operation = 'MULTIPLY'
    try:
        bsdf_node.inputs['Specular IOR Level'].default_value = 0
    except KeyError:
        bsdf_node.inputs['Specular'].default_value = 0

    # Connect nodes
    node_tree.links.new(texture_node.outputs[0], separate_node.inputs[0])
    node_tree.links.new(separate_node.outputs[2], colorramp_node.inputs[0])
    # node_tree.links.new(multiply_node.outputs[0], maprange_node.inputs[0])
    # node_tree.links.new(maprange_node.outputs[0], colorramp_node.inputs[0])
    node_tree.links.new(colorramp_node.outputs[0], bsdf_node.inputs[0])
    node_tree.links.new(bsdf_node.outputs[0], output_node.inputs[0])


def add_grid(
        material,
        grid_kwargs={},
):
    ''' Insert a mix shader connected to a image node with a grid'''
    # Get material node tree and nodes
    node_tree = material.node_tree
    nodes = node_tree.nodes

    image_node = nodes.new('ShaderNodeTexImage')
    image_node.location = (250, 400)

    mix_shader_node = nodes.new('ShaderNodeMixShader')
    mix_shader_node.location = (500, 400)

    # get output node
    output_node = nodes["Material Output"]

    # Get node which is connected to output.
    for link in node_tree.links:
        if link.to_node == output_node:
            second_to_last_node = link.from_node
            node_tree.links.remove(link)

    # Generate grid
    from utils.plots import generate_line_grid
    arr = generate_line_grid(**grid_kwargs)
    height, width, _ = arr.shape

    # Create empty Blender image
    img = bpy.data.images.new('image_name', width=width, height=height, alpha=True)

    # Flatten RGBA values into 1D list
    flat = arr.flatten()
    img.pixels = flat.tolist()  # assign pixels

    # image_node.image = bpy.data.images.load(grid_file)
    image_node.image = img

    # Make new links
    node_tree.links.new(image_node.outputs[0], mix_shader_node.inputs[0])
    node_tree.links.new(second_to_last_node.outputs[0], mix_shader_node.inputs[2])
    node_tree.links.new(mix_shader_node.outputs[0], output_node.inputs[0])


def use_image_texture(
        material, image_texture_file,
        multiply=False,
):
    ''' Insert a image texture and connect to BSDF

    Args:
      multiply: Insert as multiplyRGB with current setup

    '''
    # Get material node tree and nodes
    node_tree = material.node_tree
    nodes = node_tree.nodes

    image_node = nodes.new('ShaderNodeTexImage')
    image_node.location = (0, 400)

    # get output node
    bsdf_node = nodes["Principled BSDF"]

    # Setup nodes
    image_node.image = bpy.data.images.load(image_texture_file)
    image_node.location = (200, 400)

    if multiply:
        # Get node connected to bsdf
        for link in node_tree.links:
            if link.to_node == bsdf_node:
                second_to_last_node = link.from_node
                node_tree.links.remove(link)

        mix_node = nodes.new('ShaderNodeMixRGB')
        # Setup nodes
        mix_node.inputs[0].default_value = 1.0
        mix_node.blend_type = 'MULTIPLY'
        mix_node.location = (600, 400)

        # Make new links
        node_tree.links.new(image_node.outputs[0], mix_node.inputs[1])
        node_tree.links.new(second_to_last_node.outputs[0], mix_node.inputs[2])

        node_tree.links.new(mix_node.outputs[0], bsdf_node.inputs[0])

    else:
        # Simply connect image node and bsdf node
        node_tree.links.new(image_node.outputs[0], bsdf_node.inputs[0])


def setup_world(hdri='forest.exr', hdri_level=None):
    ''' Setup world node with hdri light '''

    scene = bpy.context.scene
    node_tree = scene.world.node_tree
    nodes = node_tree.nodes

    hdri_list = [
        sl.path for sl in bpy.context.preferences.studio_lights
        if sl.type == 'WORLD']
    hdri_dict = {os.path.basename(hdri): hdri for hdri in hdri_list}

    # Clear all nodes
    nodes.clear()

    # Add Background node
    node_background = nodes.new(type='ShaderNodeBackground')
    # And possibly set the strength level
    if hdri_level is not None:
        node_background.inputs['Strength'].default_value = hdri_level

    # Add Environment Texture node
    node_environment = nodes.new('ShaderNodeTexEnvironment')
    # Load and assign the image to the node property
    hdri_file = hdri_dict[hdri]
    node_environment.image = bpy.data.images.load(hdri_file)
    node_environment.location = (-300, 0)

    # Add Output node
    node_output = nodes.new(type='ShaderNodeOutputWorld')
    node_output.location = (200, 0)

    # Add Texture coordinates node
    node_texture = nodes.new(type='ShaderNodeTexCoord')
    node_texture.location = (-700, 0)

    # Add Mapping node
    node_mapping = nodes.new(type='ShaderNodeMapping')
    node_mapping.location = (-500, 0)

    # Link all nodes
    links = node_tree.links
    links.new(node_texture.outputs[0], node_mapping.inputs[0])
    links.new(node_mapping.outputs[0], node_environment.inputs[0])
    links.new(
        node_environment.outputs["Color"], node_background.inputs["Color"])
    links.new(
        node_background.outputs["Background"], node_output.inputs["Surface"])


def setup_camera(
        resolution_x=1920,
        resolution_y=1080,
        x=0, y=0, z=10,
        rx=0, ry=0, rz=0,
        size=(50, 50),
        camera_name='Camera', camera_type='ORTHO',
):
    '''
    Setup camera to produce array
    '''
    camera = bpy.data.objects[camera_name]

    # Set resolution from array shape
    scene = bpy.data.scenes["Scene"]
    scene.render.resolution_x = resolution_x
    scene.render.resolution_y = resolution_y

    # Set location
    camera.location.x = x
    camera.location.y = y
    camera.location.z = z
    # Set camera rotation in euler angles
    camera.rotation_mode = 'XYZ'
    camera.rotation_euler[0] = rx*(np.pi/180.0)
    camera.rotation_euler[1] = ry*(np.pi/180.0)
    camera.rotation_euler[2] = rz*(np.pi/180.0)
    # Setup camera type
    camera.data.type = camera_type
    camera.data.ortho_scale = max(size)

    bpy.context.scene.render.image_settings.file_format = 'PNG'
    bpy.context.scene.render.image_settings.color_depth = '16'
    bpy.context.scene.render.image_settings.compression = 20

    return camera


def setup_top_camera(
        array, grid_size, size,
        camera_name='Camera', camera_type='ORTHO',
):
    '''
    Setup camera to produce array

    Args:
      size: physical size of camera

    '''
    camera = bpy.data.objects[camera_name]
    z = np.max(array) + 1.0  # due to camera extension?
    z += 10  # Add 10 m margin, since it doesn't matter for 'Depth'

    rx, ry, rz = 0, 0, 0
    x, y = 0, 0

    # Set resolution from array shape
    scene = bpy.data.scenes["Scene"]
    scene.render.resolution_x = grid_size[0]
    scene.render.resolution_y = grid_size[1]

    # Set location
    camera.location.x = x
    camera.location.y = y
    camera.location.z = z
    # Set camera rotation in euler angles
    camera.rotation_mode = 'XYZ'
    camera.rotation_euler[0] = rx*(np.pi/180.0)
    camera.rotation_euler[1] = ry*(np.pi/180.0)
    camera.rotation_euler[2] = rz*(np.pi/180.0)
    # Setup camera type
    camera.data.type = camera_type
    camera.data.ortho_scale = max(size)

    bpy.context.scene.render.image_settings.file_format = 'PNG'
    bpy.context.scene.render.image_settings.color_depth = '16'
    bpy.context.scene.render.image_settings.compression = 20

    return camera


def setup_angled_camera(
        angle=45,
        center=[0, 0, 0],
        distance=None,
        view_angle=50,
        camera_name='Camera',
):
    if distance is None:
        distance = 150  # replace by auto calc

    x = center[0] + 0
    y = center[1] - np.sin(np.deg2rad(angle)) * distance
    z = center[2] + np.cos(np.deg2rad(angle)) * distance

    rx = angle
    ry, rz = 0, 0

    camera = bpy.data.objects[camera_name]
    # Set location
    camera.location.x = x
    camera.location.y = y
    camera.location.z = z

    # Set camera rotation in euler angles
    camera.rotation_mode = 'XYZ'
    camera.rotation_euler[0] = rx*(np.pi/180.0)
    camera.rotation_euler[1] = ry*(np.pi/180.0)
    camera.rotation_euler[2] = rz*(np.pi/180.0)

    camera.data.lens_unit = 'FOV'
    camera.data.angle = np.deg2rad(view_angle)
    camera.data.clip_end = 500

    return camera


def set_view_to_camera():
    for area in bpy.context.screen.areas:
        if area.type == 'VIEW_3D':
            area.spaces[0].region_3d.view_perspective = 'CAMERA'
            break


def add_holdout_plane(array, name='Plane'):
    ''' Add holdout plane '''

    # Remove any object called <name> to avoid duplicate complications
    if name in bpy.data.objects:
        obj_to_remove = bpy.data.objects[name]
        bpy.data.objects.remove(obj_to_remove, do_unlink=True)

    # Add plane
    bpy.ops.mesh.primitive_plane_add(
        size=2000,
        align='WORLD',
        location=(0, 0, -100),
        rotation=(0, 0, 0),
    )
    # Create material
    mat = bpy.data.materials.get("HoldoutMaterial")
    if mat is None:
        # create material
        mat = bpy.data.materials.new(name="HoldoutMaterial")

    # Assign it to object
    ob = bpy.data.objects.get("Plane")
    if ob.data.materials:
        # assign to 1st material slot
        ob.data.materials[0] = mat
    else:
        # no slots
        ob.data.materials.append(mat)

    setup_holdout_material(mat)


def set_object_material(obj, material):
    ''' Set material on object'''
    # Get object by name if obj is string
    if type(obj) == str:
        pass
    else:
        print("type(obj):{}".format(type(obj)))

    if obj.data.materials:
        # assign to 1st material slot
        obj.data.materials[0] = material
    else:
        # no slots
        obj.data.materials.append(material)


def setup_holdout_material(material):
    # Setup to use nodes
    material.use_nodes = True

    # Get material node tree and nodes
    node_tree = material.node_tree
    nodes = node_tree.nodes

    # Remove all previous nodes
    for node in nodes:
        nodes.remove(node)

    # Create nodes
    holdout_node = nodes.new('ShaderNodeHoldout')
    output_node = nodes.new('ShaderNodeOutputMaterial')
    holdout_node.location = (-300, 0)
    output_node.location = (0, 0)

    # Link nodes
    links = node_tree.links
    links.new(holdout_node.outputs[0], output_node.inputs[0])


def setup_bsdf(material,
               specular=0.0,
               roughness=1.0,
               ):
    node_tree = material.node_tree
    nodes = node_tree.nodes

    bsdf_node = nodes["Principled BSDF"]
    bsdf_node.inputs['Specular'].default_value = specular
    bsdf_node.inputs['Roughness'].default_value = roughness


def hide_light():
    ob = bpy.data.objects.get("Light")
    ob.hide_render = True
    ob.hide_viewport = True


def setup_lights(
        use_hdri=False,
        hdri_level=1.0,
        level=None,
        sun_position=(0, 0, 70),
        sun_angle=55,
):
    setup_sun(*sun_position, sun_angle)

    if use_hdri:
        setup_world(hdri='forest.exr', hdri_level=hdri_level)
        hide_light()

    if level is not None:
        # Explicitly set sun level
        ob = bpy.data.objects.get("Light")
        ob.hide_render = False
        ob.hide_viewport = False
        ob.data.energy = level


def setup_sun(x=0, y=0, z=70, angle=55):
    # Place sun
    light = bpy.data.objects['Light']
    light.location.x = x/2
    light.location.y = y
    light.location.z = 50
    light.data.type = 'SUN'
    light.data.energy = 10.0
    # Rotate sun
    (rx, ry, rz) = (angle, 0.0, 180.0)
    light.rotation_mode = 'XYZ'
    light.rotation_euler[0] = rx*(np.pi/180.0)
    light.rotation_euler[1] = ry*(np.pi/180.0)
    light.rotation_euler[2] = rz*(np.pi/180.0)


def colormap_to_colorramp(colorramp, colormap, gamma=2.2):
    '''
    Apply a matplotlib colormap to a blender colorramp

    bpy.types.ColorRamp
    matplotlib.colors.LinearSegmentedColormap
    '''
    # import matplotlib.cm as cm
    # from matplotlib.colors import rgb2hex

    cr = colorramp

    N = 32
    # Add up to 32 elements in colorramp
    for i in range(N-len(cr.elements)):
        cr.elements.new(i/N)

    a = np.arange(N)/(N-1)
    # colormap = cm.bone
    colors = colormap(a)
    # Make list of hex
    # hexs = [rgb2hex(colors[i, :]) for i in range(colors.shape[0])]

    # Test to 'invert gamma'
    if gamma is not None:
        colors = np.power(colors, gamma)

    # Get and setup colorramp
    # Setup colorramp
    for i, e in enumerate(cr.elements):
        e.position = a[i]
        new_color = (colors[i, 0], colors[i, 1], colors[i, 2], 0)
        e.color = new_color


def grid_from_array(array, size, center=(0, 0), name='Grid', **kwargs):
    '''
    np array of 2 dim
    size: [x, y] size in m
    '''
    import bmesh

    # Remove any object called <name> to avoid duplicate complications
    if name in bpy.data.objects:
        grid_to_remove = bpy.data.objects[name]
        bpy.data.objects.remove(grid_to_remove, do_unlink=True)

    # Create blender mesh
    rows, cols = array.shape
    bpy.ops.mesh.primitive_grid_add(
        x_subdivisions=rows-1, y_subdivisions=cols-1)

    # obj = bpy.data.objects[name]
    obj = bpy.context.object
    obj.dimensions = (size[0], size[1], 0)
    # mesh = obj.data
    obj.location.x = center[0]
    obj.location.y = center[1]
    obj.name = name

    bpy.ops.object.mode_set(mode="EDIT")
    bpy.app.debug = True

    context = bpy.context
    bm = bmesh.from_edit_mesh(context.edit_object.data)
    # deselect all
    for v in bm.verts:
        v.select = False

    bm.verts.ensure_lookup_table()
    for x in range(rows):
        for y in range(cols):
            bm.verts[rows*y + x].co.z = array[x, y]

    bpy.ops.object.mode_set(mode="OBJECT")

    return obj


def get_depth():
    """Obtains depth map from Blender render.
    :return: The depth map of the rendered camera view as a numpy array of size (H,W).
    """
    for image in bpy.data.images:
        print(f"image:{image}")
    z = bpy.data.images['Viewer Node']
    # z = bpy.data.images['Render Result']
    w, h = z.size
    dmap = np.array(z.pixels[:], dtype=np.float32)  # convert to numpy array
    dmap = np.reshape(dmap, (h, w, 4))[:, :, 0]

    # Transpose first two axes
    axes = np.arange(dmap.ndim)
    axes[:2] = [1, 0]
    dmap = np.transpose(dmap, axes=axes)

    return dmap


def render(
        render_engine='CYCLES',
        cycles_samples=128,
        save_blender=False,
        filename=None,
        save_dir=None,
):
    bpy.context.scene.render.engine = render_engine
    bpy.context.scene.cycles.samples = cycles_samples

    if filename is not None:
        filename = os.path.join(save_dir, filename)

    bpy.context.scene.render.filepath = filename
    bpy.ops.render.render(write_still=True)


def render_eevee(*args, **kwargs):
    return render(render_engine='BLENDER_EEVEE', *args, **kwargs)


def setup_render_z():
    # Setup scene to use nodes
    bpy.context.scene.use_nodes = True
    # Enable render z-pass
    bpy.context.scene.view_layers["ViewLayer"].use_pass_z = True
    # Get node tree for the compositor
    tree = bpy.context.scene.node_tree

    # Prepare nodes
    node_link_list = (
        ('CompositorNodeRLayers', None, 'Depth'),
        # ('CompositorNodeNormalize', 'Value', 'Value'),
        # ('CompositorNodeComposite', 'Image', None)
        ('CompositorNodeViewer', 'Image', None)
        )

    setup_nodes(tree, node_link_list)

    # Add another branch
    node_link_list = (
        ('CompositorNodeRLayers', None, 'Depth'),
        ('CompositorNodeNormalize', 'Value', 'Value'),
        ('CompositorNodeComposite', 'Image', None)
        )
    add_nodes_branch(tree, node_link_list, y=100)


def setup_segmentation_render():
    # Connect Render Layers to 'Viewer'
    # We can not get image from 'Render Result', only 'Viewer Layer'

    # Setup scene to use nodes
    bpy.context.scene.use_nodes = True
    # Get node tree for the compositor
    tree = bpy.context.scene.node_tree

    # Prepare nodes
    node_link_list = (
        ('CompositorNodeRLayers', None, 'Image'),
        ('CompositorNodeViewer', 'Image', None)
        )
    setup_nodes(tree, node_link_list)

    # Add another branch
    node_link_list = (
        ('CompositorNodeRLayers', None, 'Image'),
        ('CompositorNodeComposite', 'Image', None)
        )
    add_nodes_branch(tree, node_link_list, y=100)

    # Set color management to use raw linear colors
    bpy.context.scene.display_settings.display_device = 'sRGB'
    bpy.context.scene.view_settings.view_transform = 'Raw'

    # Loop through all objects in the current scene, and color
    for obj in bpy.data.objects:
        # Check if 'Ground' or 'Rock' is in the object's name
        if "Ground" in obj.name:
            set_emission_material(obj, [0, 0, 0])
        if "Rock" in obj.name:
            set_emission_material(obj, [1, 1, 1])


def set_emission_material(obj, color):
    # Ensure the object is of type 'MESH'
    if obj.type == 'MESH':
        # Remove all existing materials from the object
        obj.data.materials.clear()

        # Create a new material
        mat = bpy.data.materials.new(name="EmissionMaterial")

        # Use nodes for the material
        mat.use_nodes = True
        nodes = mat.node_tree.nodes

        # Clear default nodes
        nodes.clear()

        # Create Emission node
        emission_node = nodes.new(type='ShaderNodeEmission')
        emission_node.location = (0, 0)
        emission_node.inputs[0].default_value = (color[0], color[1], color[2], 1)  # RGB and Alpha

        # Create Material Output node
        output_node = nodes.new(type='ShaderNodeOutputMaterial')
        output_node.location = (200, 0)

        # Link Emission node to Material Output
        links = mat.node_tree.links
        links.new(emission_node.outputs[0], output_node.inputs[0])

        # Assign material to object
        obj.data.materials.append(mat)
    else:
        print("The function is only applicable to mesh objects.")


def setup_nodes(
        node_tree,
        node_link_list,
        separation_x=50,
        remove_old=True,
):
    '''
    Create nodes and links from a list of node-link info

    node_link_list = (
        ('CompositorNodeRLayers', None, 'Depth'),
        ('CompositorNodeNormalize', 'Value', 'Value'),
        ('CompositorNodeComposite', 'Image', None)
        )
    '''
    if remove_old:
        # Remove old nodes
        for node in node_tree.nodes:
            node_tree.nodes.remove(node)
            print(f"Remove:{node}")

    previous_output = None
    accumulated_width = 0

    # Loop and create nodes and links
    for node_name, input_name, output_name in node_link_list:
        new_node = node_tree.nodes.new(node_name)
        print(f"Add:{new_node}")

        # Place node
        new_node.location = (accumulated_width, 0)
        accumulated_width += new_node.width + separation_x

        # Link to previous output, if it exists
        if previous_output is not None:
            node_tree.links.new(previous_output, new_node.inputs[input_name])
        # Store output, if it exist
        if output_name is not None:
            previous_output = new_node.outputs[output_name]


def add_nodes_branch(
        node_tree,
        node_link_list,
        separation_x=50,
        y=0
):
    # Find first node-type in node_link_list in the node tree
    node_name, input_name, output_name = node_link_list[0]
    for i, node in enumerate(node_tree.nodes):
        if node_name == type(node).__name__:
            break

    # Setup location and the previous output
    accumulated_width = node.location[0] + node.width + separation_x
    previous_output = node.outputs[output_name]

    # Loop and create nodes and links
    # (skipping the first entry, as this already exists)
    for node_name, input_name, output_name in node_link_list[1:]:
        new_node = node_tree.nodes.new(node_name)
        print(f"Add:{new_node}")

        # Place node
        new_node.location = (accumulated_width, y)
        accumulated_width += new_node.width + separation_x

        # Link to previous output, if it exists
        if previous_output is not None:
            node_tree.links.new(previous_output, new_node.inputs[input_name])
        # Store output, if it exist
        if output_name is not None:
            previous_output = new_node.outputs[output_name]


def camera_to_follow_cube(cube):
    bpy.ops.object.camera_add(location=(0, 0, 0))
    camera = bpy.context.object
    # Rename the camera object if desired
    camera.name = "Camera_cube"

    camera.data.lens = 12
    camera.location = cube.location
    camera.rotation_euler = (np.deg2rad(55), 0, 0)
    # camera.location.z += 2

    # Hide cube
    cube.hide_viewport = True
    cube.hide_render = True

    # Parent camera to cube
    camera.parent = cube
    camera.parent_type = 'OBJECT'

    bpy.context.scene.camera = camera


def camera_to_follow_path():
    # Add a camera object to the scene
    bpy.ops.object.camera_add(location=(0, 0, 0))
    camera = bpy.context.object
    # Rename the camera object if desired
    camera.name = "MyCamera"

    # Set the camera's properties
    camera.data.type = 'PERSP'  # Perspective camera
    camera.data.angle = 0.8  # Field of view angle (in radians)
    camera.data.lens = 50  # Focal length of the lens

    # Set up the Follow Path constraint for the camera
    camera.constraints.new(type='FOLLOW_PATH')
    follow_path_constraint = camera.constraints['Follow Path']
    path = bpy.data.objects['Path']
    follow_path_constraint.target = path
    follow_path_constraint.up_axis = 'UP_Z'
    # follow_path_constraint.use_fixed_location = True
    # follow_path_constraint.use_fixed_rotation = True
    bpy.ops.constraint.followpath_path_animate(
        constraint="Follow Path", owner='OBJECT')

    # Create a cube object
    bpy.ops.mesh.primitive_cube_add(size=1, location=(0, 0, 0))
    cube = bpy.context.object
    cube.constraints.new(type='FOLLOW_PATH')
    cube_path_constraint = cube.constraints['Follow Path']
    cube_path_constraint.offset = -20
    cube_path_constraint.target = path

    # bpy.ops.object.constraint_add(type='TRACK_TO')
    # bpy.context.object.constraints["Track To"].target = bpy.data.objects["Cube"]
    camera.constraints.new(type='TRACK_TO')
    track_to = camera.constraints['Track To']
    track_to.target = cube

    path.data.twist_mode = 'Z_UP'
    path.data.path_duration = 1000


def image_from_terrain(terrain, name="TerrainHeight"):
    """
    Create a Blender IMAGE datablock from terrain.array, as a float buffer.

    terrain.array is assumed to be 2D (nx, ny).
    We use a grayscale image: R=G=B=normalized height, A=1.
    """
    arr = np.asarray(terrain.array, dtype=np.float32)

    # Optional: match your imshow behavior (you used terrain.array.T there)
    # If your orientation looks wrong in Blender, try arr = arr.T or np.flipud/np.fliplr.
    # arr = arr.T  # to be consistent with your plot_terrain
    arr = np.flip(arr, axis=0)

    h, w = arr.shape  # rows, cols

    # Normalize to 0..1 (or keep raw; for now we normalize for convenience)
    vmin = float(arr.min())
    vmax = float(arr.max())
    if vmax > vmin:
        norm = (arr - vmin) / (vmax - vmin)
    else:
        norm = np.zeros_like(arr, dtype=np.float32)

    # Remove old image with same name if it exists
    if name in bpy.data.images:
        bpy.data.images.remove(bpy.data.images[name], do_unlink=True)

    img = bpy.data.images.new(
        name=name,
        width=w,
        height=h,
        alpha=True,
        float_buffer=True,   # 32-bit float per channel
    )

    # Blender expects flat [R,G,B,A, R,G,B,A, ...] in row-major order
    pixels = np.zeros((h * w, 4), dtype=np.float32)
    g = norm.ravel()
    pixels[:, 0] = g  # R
    pixels[:, 1] = g  # G
    pixels[:, 2] = g  # B
    pixels[:, 3] = 1.0  # A

    img.pixels = pixels.ravel().tolist()
    img.update()

    # Return image and also the original min/max so you can map “strength” if needed
    return img, vmin, vmax


def create_displaced_ground_from_terrain(
    terrain,
    base_cuts=128,        # resolution of manual grid
    simple_levels=2,      # extra Simple subdiv levels (0..6)
    name="Terrain",
    **kwargs,
):
    """
    Create a Blender terrain object from a terrain object
    (with .array and .extent), using a float image + Displace.

    - base_cuts: number of quads per side in the base grid
    - simple_levels: extra Simple subdivision (0–6)
    - strength: displacement strength in Blender units
    """

    strength = np.max(terrain.array) - np.min(terrain.array)

    # Build float image directly from terrain.array
    img_name = f"{name}_height"
    img, vmin, vmax = image_from_terrain(terrain, name=img_name)

    # Remove existing object with this name
    if name in bpy.data.objects:
        obj_old = bpy.data.objects[name]
        bpy.data.objects.remove(obj_old, do_unlink=True)

    # --- Create a grid base mesh ---
    bpy.ops.mesh.primitive_grid_add(
        x_subdivisions=base_cuts,
        y_subdivisions=base_cuts,
        size=1.0,
    )
    obj = bpy.context.object
    obj.name = name

    # Scale to world size
    obj.scale.x = terrain.size[0]
    obj.scale.y = terrain.size[1]

    bpy.context.view_layer.objects.active = obj

    # Basic UVs (grid usually already has UVs, but we ensure something exists)
    bpy.ops.object.mode_set(mode='EDIT')
    bpy.ops.uv.smart_project()
    bpy.ops.object.mode_set(mode='OBJECT')

    # --- Optional Simple subdiv modifier ---
    if simple_levels > 0:
        subsurf = obj.modifiers.new(name="SubdivSimple", type='SUBSURF')
        subsurf.subdivision_type = 'SIMPLE'
        subsurf.levels = min(simple_levels, 6)
        subsurf.render_levels = min(simple_levels, 6)

    # --- Create texture from image for Displace ---
    tex = bpy.data.textures.new(name=f"{name}_height_tex", type='IMAGE')
    tex.image = img

    disp = obj.modifiers.new(name="Displace", type='DISPLACE')
    disp.texture = tex
    disp.texture_coords = 'UV'
    disp.strength = strength
    disp.mid_level = 0.0  # since we normalized to 0..1; tweak as needed

    # Apply object scale so displacement is in world units cleanly
    bpy.ops.object.transform_apply(location=False, rotation=False, scale=True)

    return obj



import os
import time
import functools
from utils.debug import debug


def debug_decorator(func):
    '''
    Wrapper around calls to print some debug info
    and measure execution time.
    '''
    @functools.wraps(func)
    def wrapper(self, call_number=None, call_total=None, *args, **kwargs):
        if call_number is not None and call_total is not None:
            self.logger.info(f"Run {self.name} {call_number+1}/{call_total}")

        if call_total is not None and call_total > 1:
            self.file_id = f'_{call_number:05d}'
        else:
            self.file_id = ''

        # Possibly debug input
        debug(self, kwargs, call_number, call_total, 'input')

        # ---- timing starts here ----
        start = time.perf_counter()

        result = func(
            self, *args,
            call_number=call_number,
            call_total=call_total,
            **kwargs)

        elapsed = time.perf_counter() - start
        # ---- timing ends here ----

        # Log timing
        self.logger.info(f"  took {elapsed:.3f} s")

        # Possibly debug output
        debug(self, result, call_number, call_total, 'output')

        return result

    return wrapper


class Module():
    create_folder = True
    '''
    Args:
    '''
    def __init__(self, save_dir=None, logger=None):
        self.save_dir = save_dir
        self.logger = logger

        self.save_dir_original = save_dir

        self.save_dir = os.path.join(self.save_dir, self.__class__.__name__)
        if self.create_folder:
            if not os.path.isdir(self.save_dir):
                os.makedirs(self.save_dir)

    def __call__(
            self, **kwargs):
        # Return kwargs which are not 'input'/'data'
        return kwargs

    def info(self, logging_string):
        ''' 'Shorter' method for logger.info, with indent. '''
        self.logger.info(f"  {logging_string}")

    def debug(self, logging_string):
        ''' 'Shorter' method for logger.debug, with indent. '''
        self.logger.debug(f"  {logging_string}")

    def start(self, **kwargs):
        # Store kwargs
        self.kwargs = kwargs
        # Initialize loop generator
        self.loop_generator_instance = self.loop_generator()

    @property
    def name(self):
        return self.__class__.__name__

    def loop_generator(self):
        """
        Default-beteende: returnera bara en gång, samma som __call_
        """
        yield self.__call__(**self.kwargs)

    def __iter__(self):
        """
        Gör modulen itererbar
        """
        return self.loop_generator_instance




# FIG1 = [
#     ('Set', "folder=basic"),
#     ('Basic', None),
#     ('Plot', None),
#     ('Save', None),
# 3]

FIG1b = [
    ('Set', "folder=basic"),
    ('Load', 'basic/Save'),
    ('WeightedSum', None),
    ('Ground', None),
    ('Camera', 75),
    ('Render', None),
]

# Basic
blender --python generate_terrain.py -- --save-dir basic --modules Basic Plot Save
blender --python generate_terrain.py -- --save-dir basic --modules Load:basic/Save WeightedSum:[10,3,0.3] Ground Camera:75 Render:b.png

# Octaves
python generate_terrain.py --save-dir octaves --modules Octaves Plot Save
blender --python generate_terrain.py -- --save-dir octaves --modules Load:octaves/Save Random:weights WeightedSum Save:Save2 Ground Camera:75 Holdout Render
blender --python generate_terrain.py -- --save-dir octaves --modules Load:octaves/Save Random:weights WeightedSum Ground Camera:75 Holdout Render:b.png

# Rocks
python generate_terrain.py --save-dir rocks --modules Rocks Plot Save
blender --python generate_terrain.py -- --save-dir rocks --modules Load:rocks/Save Combine Ground Camera:75 Render
blender --python generate_terrain.py -- --save-dir rocks --modules Load:rocks/Save Combine Load:octaves/Save2 Combine Ground Camera:75 Render:with_terrain.png

# Holes
python generate_terrain.py --save-dir holes --modules Holes Plot Save
blender --python generate_terrain.py -- --save-dir holes --modules Load:holes/Save Combine Ground Camera:75 Render
blender --python generate_terrain.py -- --save-dir holes --modules Load:holes/Save Combine Load:octaves/Save2 Combine Ground Camera:75 Render:with_terrain.png

# Generating functions
python generate_terrain.py --save-dir gen_func --modules Gaussian Step Donut Plane Sphere Cube SmoothStep Sine Plot Save

# Generating functions 2
blender --python generate_terrain.py -- --save-dir gen_fun_2 --modules Random:10 Gaussian Combine Ground Camera:75 Render
blender --python generate_terrain.py -- --save-dir gen_fun_2 --modules SetDistribution:'height=uniform(-5,5)' Random:10 Gaussian Combine Ground Camera:75 Render:b.png

# Distributions
blender --python generate_terrain.py -- --save-dir 2Dsampling --modules Donut:"dict(width=40)" AsProbability Random:10 Sphere Combine Ground Camera:75 Render
blender --python generate_terrain.py -- --save-dir 2D_lookup --modules Plane:"dict(pitch_deg=10)" AsLookupFor:height Random:10 SaveObstacles Gaussian Combine Ground Camera:75 Render:lookup.png
python generate_terrain.py --save-dir plot_obstacles --modules LoadObstacles:2D_lookup/SaveObstacles/obstacles.npz PlotObstacles

# Function
blender --python generate_terrain.py -- --save-dir function --modules Function:'5*(x/np.max(x))**2+np.sin(y/2)' Ground Camera:65 Render

# Blender
blender --python generate_terrain.py -- --save-dir blender --modules Load:octaves/Save LoadObstacles:2D_lookup/SaveObstacles/obstacles.npz AddMeshObjects Ground Camera:75 Render Camera:top Depth Save
blender --python generate_terrain.py -- --save-dir blender --modules Load:blender/Save Ground Camera:75 Render:combined.png

# Blender 2
blender --python generate_terrain.py -- --save-dir blender --modules Load:blender/Save Ground ImageTexture:plot_obstacles/PlotObstacles/obstacles_0.png Camera:75 Render:texture1.png
blender --python generate_terrain.py -- --save-dir blender --modules Load:octaves/Save LoadObstacles:2D_lookup/SaveObstacles/obstacles.npz AddMeshObjects Ground Camera:top RenderSegmentation:segmentation_top.png
blender --python generate_terrain.py -- --save-dir blender --modules Load:blender/Save Ground ImageTexture:blender/RenderSegmentation/segmentation_top.png Camera:75 Render:texture2.png

# Loop figures
blender --python generate_terrain.py -- --save-dir loop --modules Loop:2x2 Basic WeightedSum Ground EndLoop Camera:75 Render:1.png
blender --python generate_terrain.py -- --save-dir loop --modules Loop:2x2 Seed:2 Basic WeightedSum Ground EndLoop Camera:75 Render:2.png
blender --python generate_terrain.py -- --save-dir loop --modules Loop:2x2 Seed:2 Basic WeightedSum EndLoop Stack Ground Camera:75 Render:3.png

# Random and Loop
# blender --python generate_terrain.py -- --save-dir loop --modules Loop:2x2 Random:5 Gaussian Combine Ground EndLoop Camera:75 Render:4.png
# blender --python generate_terrain.py -- --save-dir loop --modules Random:20 Loop:2x2 Gaussian Combine Ground EndLoop Camera:75 Render:5.png

# Roughness and slop
python generate_terrain.py --save-dir slope_plot --modules Loop:persistence=linspace[0,0.99,100] Octaves:"dict(random_amp=0.0)" WeightedSum Slope Roughness LogData
python generate_terrain.py --save-dir slope_plot --modules LoadData:slope_plot/LogData/data.npz PlotScatter

# Loop over Octaves parameters
python generate_terrain.py --save-dir slope_plot2 --modules Loop:random_amp=linspace[0,0.5,6] Loop:persistance=linspace[0.5,0.7,21] Octaves WeightedSum Slope Roughness LogData
python generate_terrain.py --save-dir slope_plot2 --modules LoadData:slope_plot2/LogData/data.npz PlotScatter:"dict(color='random_amp',cmap='cet_bjy')"

# 500 Octaves terrains
python generate_terrain.py --save-dir proxy-roughness --modules Loop:num_octaves=arange[1,11] Loop:50 Octaves Roughness CombineRoughness WeightedSum Roughness LogData
python generate_terrain.py --save-dir proxy-roughness --modules LoadData:proxy-roughness/LogData/data.npz PlotScatter:"dict(color='num_octaves',cmap='cet_bjy',grid=True)"

# Roughness and slope grid
python generate_terrain.py --save-dir slope_grid --modules Octaves Save
blender --python generate_terrain.py -- --save-dir slope_grid --modules Loop:target_slope_deg=linspace[5,20,4] Loop:target_roughness=linspace[1.01,1.07,4] Load:slope_grid/Save Octaves:"dict(random_amp=0.0,random_sign=False,only_generate_weights=True)" Slope SetSlope Roughness SetRoughness WeightedSum Ground Camera:65 Render ClearScene ClearTerrain

# Number of rocks
python generate_terrain.py --save-dir rocksize_fraction --module Resolution:10 Loop:fraction=linspace[0.5,0.95,10] Loop:rock_size=linspace[0.5,4,15] Rocks FindRocks SurfaceStructure LogData
python generate_terrain.py --save-dir rocksize_fraction --module LoadData PlotScatter:"dict(color='fraction',cmap='cet_bjy',grid=True)"

# FindRocks and image texture
blender --python generate_terrain.py -- --save-dir rocksize_fraction --module Resolution:10 Loop:fraction=linspace[0.5,0.8,4] Loop:rock_size=linspace[1,4,4] Rocks FindRocks SurfaceStructure PlotObstacles ClearScene Ground ImageTexture Camera:45 Holdout Render --settings exportmode:True

# Plot height intervals
python generate_terrain.py --save-dir height_intervals --module Resolution:10 Loop:rock_size=linspace[0.5,4,200] Rocks FindRocks SurfaceStructure LogData
python generate_terrain.py --save-dir height_intervals --module LoadData PlotScatter:"dict(grid=True)"


# Spatial distribution of rocks
blender --python generate_terrain.py -- --save-dir spatial --module Resolution:10 Basic:10 Scale:10 Clip Plot AsFactor Rocks Combine Scale FindRocks SurfaceStructure PlotObstacles Ground ImageTexture Camera:65 Render --settings exportmode:True
blender --python generate_terrain.py -- --save-dir spatial --module Resolution:10 Plane Ground ImageTexture:spatial/Plot/terrain_temp_00000_0.png Camera:65 Holdout Render:mask.png

# Rock position restrictions and random_shift
python generate_terrain.py --save-dir hexagon --module Loop:1000 Rocks:"dict(fraction=0.9,rock_size=4)" EndLoop Combine:Max Plot


if __name__ == "__main__":
    """
    Script entry point for plot/renders of example configurations

    Usage:
    Run the script directly with Python:
            python -m artificial_terrains.examples

    # Note: for the reason of my probably not handling all imports
    # correctly, the following does not work:
    python artificial_terrains_cfg.py

    Or run it inside Blender in background mode:
        blender --python artificial_terrains_cfg.py --background

    Notes:
    - Rendering requires Blender and artificial_terrains package
    """
    current_module = globals()
    configs = {
        name: value for name, value in current_module.items()
        if name.isupper() and isinstance(value, list)
    }

    for name, config in configs.items():
        print(f"\n=== Running config: {name} ===")
        import artificial_terrains as at

        base_cfg = [
            ('Size', [30, 30]),
            ('GridSize', [300, 300]),
        ]

        plot_cfg = [
            ('Plot', {'filename_base': f'{name}',
                      'folder': 'example_plots'}),
        ]

        render_cfg = [
            # ('ClearScene', None),
            ('BasicSetup', None),
            ('Holdout', None),
        ]

        # Run script for plots
        at.run(base_cfg + config, verbose=True)

        # Try to run as blender script
        try:
            at.run(base_cfg + config + render_cfg, verbose=True)
        except ModuleNotFoundError:
            pass
